# coding utf-8
import pytest

from ladybug_comfort.collection.adaptive import Adaptive, PrevailingTemperature
from ladybug_comfort.parameter.adaptive import AdaptiveParameter

from ladybug_comfort.adaptive import adaptive_comfort_ashrae55, \
    adaptive_comfort_en15251, adaptive_comfort_conditioned, \
    cooling_effect_ashrae55, cooling_effect_en15251, t_operative, \
    ashrae55_neutral_offset_from_ppd, en15251_neutral_offset_from_comfort_class, \
    weighted_running_mean_hourly, weighted_running_mean_daily, \
    check_prevailing_temperatures_ashrae55, check_prevailing_temperatures_en15251

from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.datacollection import HourlyContinuousCollection, DailyCollection, \
    MonthlyCollection, MonthlyPerHourCollection
from ladybug.epw import EPW

from ladybug.datatype.temperature import Temperature, PrevailingOutdoorTemperature
from ladybug.datatype.speed import AirSpeed


def test_t_operative():
    """Test the t_operative function"""
    op_temp = t_operative(22, 28)
    assert op_temp == 25


def test_adaptive_comfort_ashrae55():
    """Test the adaptive_comfort_ashrae55 function"""
    # test typical condition
    comf_result = adaptive_comfort_ashrae55(22, 25)
    assert comf_result['to'] == 25
    assert comf_result['t_comf'] == pytest.approx(24.62, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(0.3799, rel=1e-2)
    assert comf_result['deg_comf'] == comf_result['to'] - comf_result['t_comf']

    # test a cooler outdoor case
    comf_result = adaptive_comfort_ashrae55(16, 25)
    assert comf_result['to'] == 25
    assert comf_result['t_comf'] == pytest.approx(22.76, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(2.23999, rel=1e-2)
    assert comf_result['deg_comf'] == comf_result['to'] - comf_result['t_comf']

    # test a very cold outdoor case
    comf_result = adaptive_comfort_ashrae55(5, 23)
    assert comf_result['to'] == 23
    assert comf_result['t_comf'] == pytest.approx(20.900, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(2.0999, rel=1e-2)

    # testa a very hot outdoor case
    comf_result = adaptive_comfort_ashrae55(35, 28)
    assert comf_result['to'] == 28
    assert comf_result['t_comf'] == pytest.approx(28.185, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-0.185, rel=1e-2)

    # test a fully conditioned case
    comf_result = adaptive_comfort_conditioned(24, 23, 1, 'ASHRAE-55')
    assert comf_result['to'] == 23
    assert comf_result['t_comf'] == pytest.approx(24.76, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-1.76, rel=1e-2)

    # test a partially conditioned case
    comf_result = adaptive_comfort_conditioned(24, 23, 0.5, 'ASHRAE-55')
    assert comf_result['to'] == 23
    assert comf_result['t_comf'] == pytest.approx(25.0, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-2.0, rel=1e-2)

    # test a air speed cooling effect function
    assert cooling_effect_ashrae55(1.5, 26) == 2.2
    assert cooling_effect_ashrae55(1.0, 26) == 1.8
    assert cooling_effect_ashrae55(0.7, 26) == 1.2


def test_adaptive_comfort_en15251():
    """Test the adaptive_comfort_en15251 function"""
    # test typical condition
    comf_result = adaptive_comfort_en15251(22, 25)
    assert comf_result['to'] == 25
    assert comf_result['t_comf'] == pytest.approx(26.06, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-1.06, rel=1e-2)
    assert comf_result['deg_comf'] == comf_result['to'] - comf_result['t_comf']

    # test a slightly cool case
    comf_result = adaptive_comfort_en15251(16, 25)
    assert comf_result['to'] == 25
    assert comf_result['t_comf'] == pytest.approx(24.08, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(0.9199, rel=1e-2)
    assert comf_result['deg_comf'] == comf_result['to'] - comf_result['t_comf']

    # test a very cold outdoor case
    comf_result = adaptive_comfort_en15251(5, 23)
    assert comf_result['to'] == 23
    assert comf_result['t_comf'] == pytest.approx(22.1, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(0.8999, rel=1e-2)

    # testa a very hot outdoor case
    comf_result = adaptive_comfort_en15251(35, 28)
    assert comf_result['to'] == 28
    assert comf_result['t_comf'] == pytest.approx(28.7, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-0.7, rel=1e-2)

    # test a fully conditioned case
    comf_result = adaptive_comfort_conditioned(24, 23, 1, 'EN-15251')
    assert comf_result['to'] == 23
    assert comf_result['t_comf'] == pytest.approx(24.76, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-1.76, rel=1e-2)

    # test a partially conditioned case
    comf_result = adaptive_comfort_conditioned(24, 23, 0.5, 'EN-15251')
    assert comf_result['to'] == 23
    assert comf_result['t_comf'] == pytest.approx(25.74, rel=1e-2)
    assert comf_result['deg_comf'] == pytest.approx(-2.74, rel=1e-2)

    # test a air speed cooling effect function
    assert cooling_effect_en15251(1.5, 26) == pytest.approx(3.707498, rel=1e-2)
    assert cooling_effect_en15251(1.0, 26) == pytest.approx(2.9835, rel=1e-2)
    assert cooling_effect_en15251(0.7, 26) == pytest.approx(2.34662122, rel=1e-2)


def test_ashrae55_neutral_offset_from_ppd():
    """Test the ashrae55_neutral_offset_from_ppd function."""
    assert ashrae55_neutral_offset_from_ppd(90) == 2.5
    assert ashrae55_neutral_offset_from_ppd(80) == 3.5
    with pytest.raises(Exception):
        ashrae55_neutral_offset_from_ppd(110)
    with pytest.raises(Exception):
        ashrae55_neutral_offset_from_ppd(-10)


def test_en15251_neutral_offset_from_comfort_class():
    """Test the ashrae55_neutral_offset_from_ppd function."""
    assert en15251_neutral_offset_from_comfort_class(1) == 2
    assert en15251_neutral_offset_from_comfort_class(2) == 3
    assert en15251_neutral_offset_from_comfort_class(3) == 4
    with pytest.raises(Exception):
        en15251_neutral_offset_from_comfort_class(0)
    with pytest.raises(Exception):
        en15251_neutral_offset_from_comfort_class(4)


def test_weighted_running_mean_hourly():
    """Test the weighted_running_mean_hourly function."""
    # Test with typical values
    outdoor = list(range(24)) * 365
    prevailing = weighted_running_mean_hourly(outdoor)
    for temp in prevailing[:24]:
        assert temp == prevailing[0]
    assert prevailing[0] == pytest.approx(11.5, rel=1e-2)

    # test with a number of values not divisible by 24
    outdoor = list(range(25)) * 7
    prevailing = weighted_running_mean_hourly(outdoor)
    for temp in prevailing[:24]:
        assert temp == prevailing[0]
    assert prevailing[0] == pytest.approx(12.42215, rel=1e-2)

    # test that an exception is thrown when values are less than a week.
    outdoor = list(range(24)) * 5
    with pytest.raises(Exception):
        prevailing = weighted_running_mean_hourly(outdoor)


def test_weighted_running_mean_daily():
    """Test the weighted_running_mean_daily function."""
    # Test with typical values
    outdoor = list(range(365))
    prevailing = weighted_running_mean_daily(outdoor)
    assert prevailing[0] == pytest.approx(362.1316, rel=1e-2)

    # test that an exception is thrown when values are less than a week.
    outdoor = list(range(5))
    with pytest.raises(Exception):
        prevailing = weighted_running_mean_daily(outdoor)


def test_check_prevailing_temperatures_ashrae55():
    """Test the check_prevailing_temperatures_ashrae55 function."""
    prev_temps = [22] * 24
    all_in_range, msg = check_prevailing_temperatures_ashrae55(prev_temps)
    assert all_in_range is True
    assert msg.startswith('All')

    prev_temps = range(30)
    all_in_range, msg = check_prevailing_temperatures_ashrae55(prev_temps)
    assert all_in_range is False
    assert msg.startswith('10')

    prev_temps = range(18, 40)
    all_in_range, msg = check_prevailing_temperatures_ashrae55(prev_temps)
    assert all_in_range is False
    assert msg.startswith('6')

    prev_temps = range(50)
    all_in_range, msg = check_prevailing_temperatures_ashrae55(prev_temps)
    assert all_in_range is False
    assert msg.startswith('10')


def test_check_prevailing_temperatures_en15251():
    """Test the check_prevailing_temperatures_ashrae55 function."""
    prev_temps = [22] * 24
    all_in_range, msg = check_prevailing_temperatures_en15251(prev_temps)
    assert all_in_range is True
    assert msg.startswith('All')

    prev_temps = range(30)
    all_in_range, msg = check_prevailing_temperatures_en15251(prev_temps)
    assert all_in_range is False
    assert msg.startswith('10')

    prev_temps = range(18, 38)
    all_in_range, msg = check_prevailing_temperatures_en15251(prev_temps)
    assert all_in_range is False
    assert msg.startswith('7')

    prev_temps = range(50)
    all_in_range, msg = check_prevailing_temperatures_en15251(prev_temps)
    assert all_in_range is False
    assert msg.startswith('10')


def test_adaptive_parameter_init():
    """Test the initialization of the AdaptiveParameter object."""
    ashrae55_or_en15251 = False
    neutral_offset = 2
    avg_month_or_run = True
    discr_or_cont = True
    cold_prevail_temp_limit = 18
    conditioning = 0.5
    adaptive_par = AdaptiveParameter(ashrae55_or_en15251=ashrae55_or_en15251,
                                     neutral_offset=neutral_offset,
                                     avg_month_or_running_mean=avg_month_or_run,
                                     discrete_or_continuous_air_speed=discr_or_cont,
                                     cold_prevail_temp_limit=cold_prevail_temp_limit,
                                     conditioning=conditioning)
    assert adaptive_par.ashrae55_or_en15251 == ashrae55_or_en15251
    assert adaptive_par.neutral_offset == neutral_offset
    assert adaptive_par.avg_month_or_running_mean is avg_month_or_run
    assert adaptive_par.discrete_or_continuous_air_speed == discr_or_cont
    assert adaptive_par.cold_prevail_temp_limit == cold_prevail_temp_limit
    assert adaptive_par.conditioning == conditioning


def test_adaptive_parameter_default_ahsrae55():
    """Test the default AdaptiveParameter properties."""
    adaptive_par = AdaptiveParameter()
    str(adaptive_par)  # test casting the parameters to a string
    assert adaptive_par.ashrae55_or_en15251 is True
    assert adaptive_par.neutral_offset == 2.5
    assert adaptive_par.avg_month_or_running_mean is True
    assert adaptive_par.discrete_or_continuous_air_speed is True
    assert adaptive_par.cold_prevail_temp_limit == 10
    assert adaptive_par.conditioning == 0
    assert adaptive_par.standard == 'ASHRAE-55'
    assert adaptive_par.prevailing_temperature_method == 'Averaged Monthly'
    assert adaptive_par.air_speed_method == 'Discrete'
    assert adaptive_par.minimum_operative == pytest.approx(18.4, rel=1e-2)

    adaptive_par.set_neutral_offset_from_ppd(80)
    assert adaptive_par.neutral_offset == 3.5


def test_adaptive_parameter_default_en15251():
    """Test the default AdaptiveParameter properties."""
    adaptive_par = AdaptiveParameter(False)
    str(adaptive_par)  # test casting the parameters to a string
    assert adaptive_par.ashrae55_or_en15251 is False
    assert adaptive_par.neutral_offset == 3
    assert adaptive_par.avg_month_or_running_mean is False
    assert adaptive_par.discrete_or_continuous_air_speed is False
    assert adaptive_par.cold_prevail_temp_limit == 15
    assert adaptive_par.conditioning == 0
    assert adaptive_par.standard == 'EN-15251'
    assert adaptive_par.prevailing_temperature_method == 'Running Mean'
    assert adaptive_par.air_speed_method == 'Continuous'
    assert adaptive_par.minimum_operative == pytest.approx(20.75, rel=1e-2)

    adaptive_par.set_neutral_offset_from_comfort_class(1)
    assert adaptive_par.neutral_offset == 2


def test_adaptive_parameter_incorrect():
    """Test incorrect AdaptiveParameter properties."""
    with pytest.raises(Exception):
        AdaptiveParameter(neutral_offset=-2)
    with pytest.raises(Exception):
        AdaptiveParameter(neutral_offset=12)
    with pytest.raises(Exception):
        AdaptiveParameter(cold_prevail_temp_limit=5)
    with pytest.raises(Exception):
        AdaptiveParameter(cold_prevail_temp_limit=30)
    with pytest.raises(Exception):
        AdaptiveParameter(conditioning=50)


def test_comfort_check():
    """Test comfort check on AdaptiveParameter."""
    comf_result = adaptive_comfort_ashrae55(24, 28)
    adaptive_par = AdaptiveParameter()
    comf_test = adaptive_par.is_comfortable(comf_result)
    assert comf_test == 0

    comf_result = adaptive_comfort_ashrae55(24, 28)
    adaptive_par = AdaptiveParameter()
    comf_test = adaptive_par.is_comfortable(comf_result, 3)
    assert comf_test == 1


def test_thermal_condition_check():
    """Test the thermal condition check on AdaptiveParameter."""
    comf_result = adaptive_comfort_ashrae55(24, 28)
    adaptive_par = AdaptiveParameter()
    condition_test = adaptive_par.thermal_condition(comf_result)
    assert condition_test == 1

    comf_result = adaptive_comfort_ashrae55(24, 28)
    adaptive_par = AdaptiveParameter()
    condition_test = adaptive_par.thermal_condition(comf_result, 3)
    assert condition_test == 0


def test_init_adaptive_collection():
    """Test the initialization of the Adaptive collection and basic outputs."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    adapt_obj = Adaptive(prevail_temp, op_temp)

    assert adapt_obj.comfort_model == 'Adaptive'
    assert adapt_obj.calc_length == calc_length
    str(adapt_obj)  # test that the string representaiton is ok

    assert isinstance(adapt_obj.prevailing_outdoor_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.prevailing_outdoor_temperature.values) == calc_length
    assert adapt_obj.prevailing_outdoor_temperature[0] == 22
    assert isinstance(adapt_obj.operative_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.operative_temperature.values) == calc_length
    assert adapt_obj.operative_temperature[0] == 26

    assert isinstance(adapt_obj.neutral_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.neutral_temperature.values) == calc_length
    assert adapt_obj.neutral_temperature[0] == pytest.approx(24.62, rel=1e-3)
    assert isinstance(adapt_obj.degrees_from_neutral, HourlyContinuousCollection)
    assert len(adapt_obj.degrees_from_neutral.values) == calc_length
    assert adapt_obj.degrees_from_neutral[0] == pytest.approx(1.3799, rel=1e-3)


def test_init_adaptive_collection_mrt():
    """Test the initialization of the Adaptive collection with MRT."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, [24] * calc_length)
    adapt_obj = Adaptive.from_air_and_rad_temp(prevail_temp, air_temp, 28)

    assert adapt_obj.comfort_model == 'Adaptive'
    assert adapt_obj.calc_length == calc_length
    str(adapt_obj)  # test that the string representaiton is ok

    assert isinstance(adapt_obj.prevailing_outdoor_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.prevailing_outdoor_temperature.values) == calc_length
    assert adapt_obj.prevailing_outdoor_temperature[0] == 22
    assert isinstance(adapt_obj.operative_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.operative_temperature.values) == calc_length
    assert adapt_obj.operative_temperature[0] == 26

    assert isinstance(adapt_obj.neutral_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.neutral_temperature.values) == calc_length
    assert adapt_obj.neutral_temperature[0] == pytest.approx(24.62, rel=1e-3)
    assert isinstance(adapt_obj.degrees_from_neutral, HourlyContinuousCollection)
    assert len(adapt_obj.degrees_from_neutral.values) == calc_length
    assert adapt_obj.degrees_from_neutral[0] == pytest.approx(1.3799, rel=1e-3)


def test_adaptive_collection_defaults():
    """Test the default inputs assigned to the Adaptive collection."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    adapt_obj = Adaptive(prevail_temp, op_temp)

    assert isinstance(adapt_obj.air_speed, HourlyContinuousCollection)
    assert len(adapt_obj.air_speed.values) == calc_length
    assert adapt_obj.air_speed[0] == 0.1

    assert isinstance(adapt_obj.comfort_parameter, AdaptiveParameter)
    default_par = AdaptiveParameter()
    assert adapt_obj.comfort_parameter.ashrae55_or_en15251 == default_par.ashrae55_or_en15251
    assert adapt_obj.comfort_parameter.neutral_offset == default_par.neutral_offset
    assert adapt_obj.comfort_parameter.avg_month_or_running_mean == default_par.avg_month_or_running_mean
    assert adapt_obj.comfort_parameter.discrete_or_continuous_air_speed == default_par.discrete_or_continuous_air_speed
    assert adapt_obj.comfort_parameter.cold_prevail_temp_limit == default_par.cold_prevail_temp_limit
    assert adapt_obj.comfort_parameter.conditioning == default_par.conditioning


def test_adaptive_collection_comfort_outputs():
    """Test the is_comfortable and thermal_condition outputs of the collection."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, range(20, 20 + calc_length))
    adapt_obj = Adaptive(prevail_temp, op_temp)

    assert isinstance(adapt_obj.is_comfortable, HourlyContinuousCollection)
    assert len(adapt_obj.is_comfortable.values) == calc_length
    assert adapt_obj.is_comfortable[0] == 0
    assert adapt_obj.is_comfortable[5] == 1
    assert adapt_obj.is_comfortable[10] == 0

    assert isinstance(adapt_obj.thermal_condition, HourlyContinuousCollection)
    assert len(adapt_obj.thermal_condition.values) == calc_length
    assert adapt_obj.thermal_condition[0] == -1
    assert adapt_obj.thermal_condition[5] == 0
    assert adapt_obj.thermal_condition[10] == 1


def test_adaptive_collection_cooling_effect_output():
    """Test the cooling effect output of the Adaptive collection."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    adapt_obj = Adaptive(prevail_temp, op_temp, air_speed=0.7)

    assert isinstance(adapt_obj.cooling_effect, HourlyContinuousCollection)
    assert len(adapt_obj.cooling_effect.values) == calc_length
    assert adapt_obj.cooling_effect[0] == 1.2


def test_adaptive_collection_immutability():
    """Test that the Adaptive collection is immutable."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    adapt_obj = Adaptive(prevail_temp, op_temp)

    # check that editing the original collection does not mutate the object
    op_temp[0] = 28
    assert adapt_obj.operative_temperature[0] == 26

    # check that editing collection properties does not mutate the object
    with pytest.raises(Exception):
        adapt_obj.operative_temperature[0] = 28
    with pytest.raises(Exception):
        adapt_obj.operative_temperature.values = [28] * calc_length
    with pytest.raises(Exception):
        adapt_obj.degrees_from_neutral[0] = 0.5
    with pytest.raises(Exception):
        adapt_obj.degrees_from_neutral.values = [0.5] * calc_length

    # check that properties cannot be edited directly
    with pytest.raises(Exception):
        adapt_obj.operative_temperature = op_temp
    with pytest.raises(Exception):
        adapt_obj.degrees_from_neutral = op_temp
    with pytest.raises(Exception):
        adapt_obj.comfort_parameter = AdaptiveParameter(False)


def test_init_adaptive_collection_full_input():
    """Test the initialization of the Adaptive collection will all inputs."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    custom_par = AdaptiveParameter(True, 2, False, False, 15, 0.25)
    adapt_obj = Adaptive(prevail_temp, op_temp, 0.7, custom_par)

    assert adapt_obj.operative_temperature[0] == 26
    assert adapt_obj.air_speed[0] == 0.7
    assert adapt_obj.comfort_parameter.ashrae55_or_en15251 is True
    assert adapt_obj.comfort_parameter.neutral_offset == 2
    assert adapt_obj.comfort_parameter.avg_month_or_running_mean is False
    assert adapt_obj.comfort_parameter.discrete_or_continuous_air_speed is False
    assert adapt_obj.comfort_parameter.cold_prevail_temp_limit == 15
    assert adapt_obj.comfort_parameter.conditioning == 0.25


def test_init_adaptive_collection_full_collection_input():
    """Test initialization of the Adaptive collection with inputs as collections."""
    calc_length = 24
    prevail_header = Header(PrevailingOutdoorTemperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    prevail_temp = HourlyContinuousCollection(prevail_header, [22] * calc_length)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    air_speed_header = Header(AirSpeed(), 'm/s', AnalysisPeriod(end_month=1, end_day=1))
    air_speed = HourlyContinuousCollection(air_speed_header, [0.7] * calc_length)
    adapt_obj = Adaptive(prevail_temp, op_temp, air_speed)

    assert adapt_obj.operative_temperature[0] == 26
    assert adapt_obj.air_speed[0] == 0.7


def test_init_adaptive_collection_epw():
    """Test the initialization of the Adaptive collection with EPW input."""
    calc_length = 8760
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    adapt_obj = Adaptive(epw.dry_bulb_temperature, epw.dry_bulb_temperature)

    assert len(adapt_obj.prevailing_outdoor_temperature.values) == calc_length
    assert adapt_obj.prevailing_outdoor_temperature[0] == pytest.approx(-4.648456, rel=1e-3)
    assert len(adapt_obj.operative_temperature.values) == calc_length
    assert adapt_obj.operative_temperature[0] == -6.1

    assert len(adapt_obj.neutral_temperature.values) == calc_length
    assert adapt_obj.neutral_temperature[0] == pytest.approx(20.9, rel=1e-3)
    assert len(adapt_obj.degrees_from_neutral.values) == calc_length
    assert adapt_obj.degrees_from_neutral[0] == pytest.approx(-27.0, rel=1e-1)


def test_adaptive_collection_comfort_percent_outputs():
    """Test the percent outputs of the Adaptive collection."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    adapt_obj = Adaptive(epw.dry_bulb_temperature, epw.dry_bulb_temperature)

    assert adapt_obj.percent_comfortable == pytest.approx(12.95662, rel=1e-3)
    assert adapt_obj.percent_uncomfortable == pytest.approx(87.043378, rel=1e-3)
    assert adapt_obj.percent_neutral == pytest.approx(12.95662, rel=1e-3)
    assert adapt_obj.percent_hot == pytest.approx(6.4726027, rel=1e-3)
    assert adapt_obj.percent_cold == pytest.approx(80.570776, rel=1e-3)


def test_adaptive_collection_epw_prevailing():
    """Test the percent outputs of the Adaptive collection."""
    calc_length = 24
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    op_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    op_temp = HourlyContinuousCollection(op_temp_header, [26] * calc_length)
    adapt_obj = Adaptive(epw.dry_bulb_temperature, op_temp)

    assert isinstance(adapt_obj.prevailing_outdoor_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.prevailing_outdoor_temperature.values) == calc_length
    assert adapt_obj.prevailing_outdoor_temperature[0] == pytest.approx(-4.64845637, rel=1e-3)
    assert isinstance(adapt_obj.operative_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.operative_temperature.values) == calc_length
    assert adapt_obj.operative_temperature[0] == 26

    assert isinstance(adapt_obj.neutral_temperature, HourlyContinuousCollection)
    assert len(adapt_obj.neutral_temperature.values) == calc_length
    assert adapt_obj.neutral_temperature[0] == pytest.approx(20.9, rel=1e-3)
    assert isinstance(adapt_obj.degrees_from_neutral, HourlyContinuousCollection)
    assert len(adapt_obj.degrees_from_neutral.values) == calc_length
    assert adapt_obj.degrees_from_neutral[0] == pytest.approx(5.099999, rel=1e-3)


def test_init_prevailing_temperature_hourly():
    """Test the PrevailingTemperature object with hourly inputs."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)

    prevail_obj = PrevailingTemperature(epw.dry_bulb_temperature, True)
    assert isinstance(prevail_obj.hourly_prevailing_temperature, HourlyContinuousCollection)
    assert len(prevail_obj.hourly_prevailing_temperature.values) == 8760
    assert isinstance(prevail_obj.daily_prevailing_temperature, DailyCollection)
    assert len(prevail_obj.daily_prevailing_temperature.values) == 365
    assert isinstance(prevail_obj.monthly_prevailing_temperature, MonthlyCollection)
    assert len(prevail_obj.monthly_prevailing_temperature.values) == 12
    assert isinstance(prevail_obj.monthly_per_hour_prevailing_temperature, MonthlyPerHourCollection)
    assert len(prevail_obj.monthly_per_hour_prevailing_temperature.values) == 288

    prevail_obj = PrevailingTemperature(epw.dry_bulb_temperature, False)
    assert isinstance(prevail_obj.hourly_prevailing_temperature, HourlyContinuousCollection)
    assert len(prevail_obj.hourly_prevailing_temperature.values) == 8760
    assert isinstance(prevail_obj.daily_prevailing_temperature, DailyCollection)
    assert len(prevail_obj.daily_prevailing_temperature.values) == 365
    assert isinstance(prevail_obj.monthly_prevailing_temperature, MonthlyCollection)
    assert len(prevail_obj.monthly_prevailing_temperature.values) == 12
    assert isinstance(prevail_obj.monthly_per_hour_prevailing_temperature, MonthlyPerHourCollection)
    assert len(prevail_obj.monthly_per_hour_prevailing_temperature.values) == 288


def test_init_prevailing_temperature_daily():
    """Test the PrevailingTemperature object with daily inputs."""
    outdoor_header = Header(Temperature(), 'C', AnalysisPeriod())
    outdoor_temp = DailyCollection(outdoor_header, range(365), AnalysisPeriod().doys_int)
    outdoor_temp = outdoor_temp.validate_analysis_period()

    prevail_obj = PrevailingTemperature(outdoor_temp, True)
    assert isinstance(prevail_obj.hourly_prevailing_temperature, HourlyContinuousCollection)
    assert len(prevail_obj.hourly_prevailing_temperature.values) == 8760
    assert isinstance(prevail_obj.daily_prevailing_temperature, DailyCollection)
    assert len(prevail_obj.daily_prevailing_temperature.values) == 365
    assert isinstance(prevail_obj.monthly_prevailing_temperature, MonthlyCollection)
    assert len(prevail_obj.monthly_prevailing_temperature.values) == 12
    assert isinstance(prevail_obj.monthly_per_hour_prevailing_temperature, MonthlyPerHourCollection)
    assert len(prevail_obj.monthly_per_hour_prevailing_temperature.values) == 288

    prevail_obj = PrevailingTemperature(outdoor_temp, False)
    assert isinstance(prevail_obj.hourly_prevailing_temperature, HourlyContinuousCollection)
    assert len(prevail_obj.hourly_prevailing_temperature.values) == 8760
    assert isinstance(prevail_obj.daily_prevailing_temperature, DailyCollection)
    assert len(prevail_obj.daily_prevailing_temperature.values) == 365
    assert isinstance(prevail_obj.monthly_prevailing_temperature, MonthlyCollection)
    assert len(prevail_obj.monthly_prevailing_temperature.values) == 12
    assert isinstance(prevail_obj.monthly_per_hour_prevailing_temperature, MonthlyPerHourCollection)
    assert len(prevail_obj.monthly_per_hour_prevailing_temperature.values) == 288


def test_init_prevailing_temperature_monthly():
    """Test the PrevailingTemperature object with monthly inputs."""
    outdoor_header = Header(Temperature(), 'C', AnalysisPeriod())
    outdoor_temp = MonthlyCollection(outdoor_header, range(12), AnalysisPeriod().months_int)
    outdoor_temp = outdoor_temp.validate_analysis_period()

    prevail_obj = PrevailingTemperature(outdoor_temp, True)
    assert isinstance(prevail_obj.hourly_prevailing_temperature, HourlyContinuousCollection)
    assert len(prevail_obj.hourly_prevailing_temperature.values) == 8760
    assert isinstance(prevail_obj.daily_prevailing_temperature, DailyCollection)
    assert len(prevail_obj.daily_prevailing_temperature.values) == 365
    assert isinstance(prevail_obj.monthly_prevailing_temperature, MonthlyCollection)
    assert len(prevail_obj.monthly_prevailing_temperature.values) == 12
    assert isinstance(prevail_obj.monthly_per_hour_prevailing_temperature, MonthlyPerHourCollection)
    assert len(prevail_obj.monthly_per_hour_prevailing_temperature.values) == 288

    with pytest.raises(Exception):
        prevail_obj = PrevailingTemperature(outdoor_temp, False)



# coding=utf-8
"""Utility functions for calculating Heating and Cooling Degree-Time."""
from __future__ import division


def heating_degree_time(t, t_base=18):
    """Calculate heating degree time at a single time interval.

    Args:
        t: The temperature at a given time interval.
        t_base: The base temperature below which a given time is considered
            to be in heating mode. This should be in the same units as the input
            temperature. Default is 18 Celcius, which is a common balance point for
            buildings.
    """
    if t < t_base:
        return t_base - t
    else:
        return 0


def cooling_degree_time(t, t_base=23):
    """Calculate cooling degree time at a single time interval.

    Args:
        t: The temperature at a given time interval.
        t_base: The base temperature above which a given time is considered
            to be in cooling mode. This should be in the same units as the input
            temperature. Default is 23 Celcius, which is a common balance point for
            buildings.
    """
    if t > t_base:
        return t - t_base
    else:
        return 0



# coding=utf-8
"""Functions for adjusting MRT for radiative sky exchange (including shortwave solar).

The solarcal formulas of this module are taken from the following publications:

[1] Arens, E., T. Hoyt, X. Zhou, L. Huang, H. Zhang and S. Schiavon. 2015.
Modeling the comfort effects of short-wave solar radiation indoors.
Building and Environment, 88, 3-9. http://dx.doi.org/10.1016/j.buildenv.2014.09.004
https://escholarship.org/uc/item/89m1h2dg

[2] ASHRAE Standard 55 (2017). "Thermal Environmental Conditions for Human Occupancy".

Properties:
    SOLARCAL_SPLINES: A dictionary with two keys: 'standing' and 'seated'.
        Each value for these keys is a 2D matrix of projection factors
        for human geometry.  Each row refers to an degree of azimuth and each
        colum refers to a degree of altitude.
"""
from __future__ import division

from ladybug.skymodel import calc_sky_temperature
from ladybug.futil import csv_to_num_matrix

import os
import math


def _load_solarcal_splines():
    """load the spline data that gets used in solarcal."""
    try:
        cur_dir = os.path.dirname(__file__)
        solarcal_splines = {
            'seated': csv_to_num_matrix(os.path.join(
                cur_dir, '_mannequin', 'seatedspline.csv')),
            'standing': csv_to_num_matrix(os.path.join(
                cur_dir, '_mannequin', 'standingspline.csv'))}
    except IOError:
        solarcal_splines = {}
        print('Failed to import projection factor splines from CSV.'
              '\nA simpler interoplation method for Solarcal will be used.')
    return solarcal_splines


SOLARCAL_SPLINES = _load_solarcal_splines()


def outdoor_sky_heat_exch(srfs_temp, horiz_ir, diff_horiz_solar, dir_normal_solar, alt,
                          sky_exposure=1, fract_exposed=1, floor_reflectance=0.25,
                          posture='standing', sharp=135,
                          body_absorptivity=0.7, body_emissivity=0.95):
    """Perform a full outdoor sky radiant heat exchange.

    Args:
        srfs_temp: The temperature of surfaces around the person in degrees
            Celcius. This includes the ground and any other surfaces
            blocking the view to the sky. When the temperature of these
            individual surfaces are known, the input here should be the
            average temperature of the surfaces weighted by view-factor to the human.
            When such individaal surface temperatures are unknown, the outdoor
            dry bulb temperature is typically used as a proxy.
        horiz_ir: The horizontal infrared radiation intensity from the sky in W/m2.
        diff_horiz_solar: Diffuse horizontal solar irradiance in W/m2.
        dir_normal_solar: Direct normal solar irradiance in W/m2.
        alt: The altitude of the sun in degrees [0-90].
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for outdoors in an
            open field.
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include the
            body’s self-shading; only the shading from surroundings.
            Default is 1 for a person standing in an open area.
        floor_reflectance: A number between 0 and 1 the represents the
            reflectance of the floor. Default is for 0.25 which is characteristic
            of outdoor grass or dry bare soil.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        body_absorptivity: A number between 0 and 1 representing the average
            shortwave absorptivity of the body (including clothing and skin color).
            Typical clothing values - white: 0.2, khaki: 0.57, black: 0.88
            Typical skin values - white: 0.57, brown: 0.65, black: 0.84
            Default is 0.7 for average (brown) skin and medium clothing.
        body_emissivity: A number between 0 and 1 representing the average
            longwave emissivity of the body.  Default is 0.95, which is almost
            always the case except in rare situations of wearing metalic clothing.

    Returns:
        heat_exch_result: A dictionary containing results with the following keys:
            s_erf : The shortwave effective radiant field (ERF) in W/m2.
            s_dmrt : The MRT delta as a result of shortwave irradinace in C.
            l_erf : The longwave effective radiant field (ERF) in W/m2.
            l_dmrt : The MRT delta as a result of longwave sky exchange in C.
            mrt: The final MRT expereinced as a result of sky heat excahnge in C.
    """
    # set defaults using the input parameters
    fract_efficiency = 0.696 if posture == 'seated' else 0.725

    # calculate the influence of shorwave irradiance
    if alt >= 0:
        s_flux = body_solar_flux_from_parts(diff_horiz_solar, dir_normal_solar,
                                            alt, sharp, sky_exposure,
                                            fract_exposed, floor_reflectance, posture)
        short_erf = erf_from_body_solar_flux(s_flux, body_absorptivity, body_emissivity)
        short_mrt_delta = mrt_delta_from_erf(short_erf, fract_efficiency)
    else:
        short_erf = 0
        short_mrt_delta = 0

    # calculate the influence of longwave heat exchange with the sky
    long_mrt_delta = longwave_mrt_delta_from_horiz_ir(horiz_ir, srfs_temp,
                                                      sky_exposure, body_emissivity)
    long_erf = erf_from_mrt_delta(long_mrt_delta, fract_efficiency)

    # calculate final MRT as a result of both longwave and shortwave heat exchange
    sky_adjusted_mrt = srfs_temp + short_mrt_delta + long_mrt_delta
    heat_exch_result = {
        's_erf': short_erf,
        's_dmrt': short_mrt_delta,
        'l_erf': long_erf,
        'l_dmrt': long_mrt_delta,
        'mrt': sky_adjusted_mrt
    }
    return heat_exch_result


def indoor_sky_heat_exch(longwave_mrt, diff_horiz_solar, dir_normal_solar, alt,
                         sky_exposure=1, fract_exposed=1, floor_reflectance=0.25,
                         window_transmittance=0.4, posture='seated', sharp=135,
                         body_absorptivity=0.7, body_emissivity=0.95):
    """Perform a full indoor sky radiant heat exchange.

    Args:
        longwave_mrt: The longwave mean radiant temperature (MRT) expereinced
            as a result of indoor surface temperatures in C.
        diff_horiz_solar: Diffuse horizontal solar irradiance in W/m2.
        dir_normal_solar: Direct normal solar irradiance in W/m2.
        alt: The altitude of the sun in degrees [0-90].
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for a completely glass box.
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include the
            body’s self-shading; only the shading from surroundings.
            Default is 1 for a person standing in an open area.
        floor_reflectance: A number between 0 and 1 the represents the
            reflectance of the floor. Default is for 0.25 which is characteristic
            of outdoor grass or dry bare soil.
        window_transmittance: A number between 0 and 1 that represents the broadband
            solar transmittance of the window through which the sun is coming. Such
            values tend to be slightly less than the SHGC. Values might be as low as
            0.2 and could be as high as 0.85 for a single pane of glass.
            Default is 0.4 assuming a double pane window with a relatively mild
            low-e coating.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        body_absorptivity: A number between 0 and 1 representing the average
            shortwave absorptivity of the body (including clothing and skin color).
            Typical clothing values - white: 0.2, khaki: 0.57, black: 0.88
            Typical skin values - white: 0.57, brown: 0.65, black: 0.84
            Default is 0.7 for average (brown) skin and medium clothing.
        body_emissivity: A number between 0 and 1 representing the average
            longwave emissivity of the body.  Default is 0.95, which is almost
            always the case except in rare situations of wearing metalic clothing.

    Returns:
        heat_exch_result: A dictionary containing results with the following keys:
            erf : The shortwave effective radiant field (ERF) in W/m2.
            dmrt : The MRT delta as a result of shortwave irradinace in C.
            mrt: The final MRT expereinced as a result of sky heat excahnge in C.
    """
    # set defaults using the input parameters
    fract_efficiency = 0.696 if posture == 'seated' else 0.725

    # calculate the influence of shorwave irradiance
    if alt >= 0:
        s_flux = body_solar_flux_from_parts(diff_horiz_solar, dir_normal_solar,
                                            alt, sharp, sky_exposure,
                                            fract_exposed, floor_reflectance, posture)
        s_flux = s_flux * window_transmittance
        short_erf = erf_from_body_solar_flux(s_flux, body_absorptivity, body_emissivity)
        short_mrt_delta = mrt_delta_from_erf(short_erf, fract_efficiency)
    else:
        short_erf = 0
        short_mrt_delta = 0

    # calculate final MRT
    sky_adjusted_mrt = longwave_mrt + short_mrt_delta
    heat_exch_result = {
        'erf': short_erf,
        'dmrt': short_mrt_delta,
        'mrt': sky_adjusted_mrt
    }
    return heat_exch_result


def shortwave_from_horiz_solar(longwave_mrt, diff_horiz_solar, dir_horiz_solar, alt,
                               fract_exposed=1, floor_reflectance=0.25,
                               posture='standing', sharp=135,
                               body_absorptivity=0.7, body_emissivity=0.95):
    """Perform a shortwave radiant heat exchange using horizontal solar components.

    This is useful when building a map of MRT using the direct and diffuse
    results of a Radiance study instead of the solar components directly from
    an EPW or Wea file.  Note that all input radiation components should already
    account for the amount of sky seen and solar heat reflections off of surfaces.

    Args:
        longwave_mrt: The longwave mean radiant temperature (MRT) expereinced
            as a result of indoor surface temperatures in C.
        horiz_ir: The horizontal infrared radiation intensity from the sky in W/m2.
        diff_horiz_solar: Diffuse horizontal solar irradiance in W/m2.
        dir_horiz_solar: Direct horizontal solar irradiance in W/m2.
        alt: The altitude of the sun in degrees [0-90].
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include the
            body’s self-shading; only the shading from surroundings.
            Default is 1 for a person standing in an open area.
        floor_reflectance: A number between 0 and 1 the represents the
            reflectance of the floor. Default is for 0.25 which is characteristic
            of outdoor grass or dry bare soil.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        body_absorptivity: A number between 0 and 1 representing the average
            shortwave absorptivity of the body (including clothing and skin color).
            Typical clothing values - white: 0.2, khaki: 0.57, black: 0.88
            Typical skin values - white: 0.57, brown: 0.65, black: 0.84
            Default is 0.7 for average (brown) skin and medium clothing.
        body_emissivity: A number between 0 and 1 representing the average
            longwave emissivity of the body.  Default is 0.95, which is almost
            always the case except in rare situations of wearing metalic clothing.

    Returns:
        heat_exch_result: A dictionary containing results with the following keys:
            erf : The shortwave effective radiant field (ERF) in W/m2.
            dmrt : The MRT delta as a result of shortwave irradinace in C.
            mrt: The final MRT expereinced as a result of sky heat excahnge in C.
    """
    # set defaults using the input parameters
    fract_efficiency = 0.696 if posture == 'seated' else 0.725

    # calculate the influence of shorwave irradiance
    if alt >= 0:
        s_flux = body_solar_flux_from_horiz_parts(diff_horiz_solar, dir_horiz_solar,
                                                  alt, sharp, fract_exposed,
                                                  floor_reflectance, posture)
        short_erf = erf_from_body_solar_flux(s_flux, body_absorptivity, body_emissivity)
        short_mrt_delta = mrt_delta_from_erf(short_erf, fract_efficiency)
    else:
        short_erf = 0
        short_mrt_delta = 0

    # calculate final MRT as a result of both longwave and shortwave heat exchange
    sky_adjusted_mrt = longwave_mrt + short_mrt_delta
    heat_exch_result = {
        'erf': short_erf,
        'dmrt': short_mrt_delta,
        'mrt': sky_adjusted_mrt
    }
    return heat_exch_result


def mrt_delta_from_erf(erf, fract_efficiency=0.725, rad_trans_coeff=6.012):
    """Calculate the mean radiant temperature (MRT) delta as a result of an ERF.

    Args:
        erf: A number representing the effective radiant field (ERF) on the
            person in W/m2.
        fract_efficiency: A number representing the fraction of the body
            surface exposed to radiation from the environment. This is typically
            either 0.725 for a standing or supine person or 0.696 for a seated
            person. Default is 0.725 for a standing person.
        rad_trans_coeff: A number representing the radiant heat transfer coefficient
            in (W/m2-K).  Default is 6.012, which is almost always the case.
    """
    return erf / (fract_efficiency * rad_trans_coeff)


def erf_from_mrt_delta(mrt_delta, fract_efficiency=0.725, rad_trans_coeff=6.012):
    """Calculate the effective radiant field (ERF) from a MRT delta.

    Args:
        mrt_delta: A mean radiant temperature (MRT) delta in Kelvin or degrees Celcius.
        fract_efficiency: A number representing the fraction of the body
            surface exposed to radiation from the environment. This is typically
            either 0.725 for a standing or supine person or 0.696 for a seated
            person. Default is 0.725 for a standing person.
        rad_trans_coeff: A number representing the radiant heat transfer coefficient
            in (W/m2-K).  Default is 6.012, which is almost always the case.
    """
    return mrt_delta * fract_efficiency * rad_trans_coeff


def longwave_mrt_delta_from_horiz_ir(horiz_ir, srfs_temp, sky_exposure=1,
                                     body_emissivity=0.95):
    """Calculate the MRT delta as a result of longwave radiant exchange with the sky.

    Note that this value is typically negative since the earth (and humans)
    tend to radiate heat out to space in the longwave portion of the spectrum.

    Args:
        horiz_ir: A float value that represents the downwelling horizontal
            infrared radiation intensity in W/m2.
        srfs_temp: The temperature of surfaces around the person in degrees
            Celcius. This includes the ground and any other surfaces
            blocking the view to the sky. Typically, the dry bulb temperature
            is used when such surface temperatures are unknown.
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for outdoors in an
            open field.
    """
    sky_temp = calc_sky_temperature(horiz_ir, body_emissivity)
    return longwave_mrt_delta_from_sky_temp(sky_temp, srfs_temp, sky_exposure)


def longwave_mrt_delta_from_sky_temp(sky_temp, srfs_temp, sky_exposure=1):
    """Calculate the MRT delta as a result of longwave radiant exchange with the sky.

    Note that this value is typically negative since the earth (and humans)
    tend to radiate heat out to space in the longwave portion of the spectrum.

    Args:
        sky_temp: The sky temperature in degrees Celcius.
        srfs_temp: The temperature of surfaces around the person in degrees
            Celcius. This includes the ground and any other surfaces
            blocking the view to the sky. Typically, the dry bulb temperature
            is used when such surface temperatures are unknown.
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for outdoors in an
            open field.
    """
    return 0.5 * sky_exposure * (sky_temp - srfs_temp)


def erf_from_body_solar_flux(solar_flux, body_absorptivity=0.7, body_emissivity=0.95):
    """Calculate effective radiant field (ERF) from incident solar flux on body in W/m2.

    Args:
        solar_flux: A number for the average solar flux over the human body in W/m2.
        body_absorptivity: A number between 0 and 1 representing the average
            shortwave absorptivity of the body (including clothing and skin color).
            Typical clothing values - white: 0.2, khaki: 0.57, black: 0.88
            Typical skin values - white: 0.57, brown: 0.65, black: 0.84
            Default is 0.7 for average (brown) skin and medium clothing.
        body_emissivity: A number between 0 and 1 representing the average
            longwave emissivity of the body.  Default is 0.95, which is almost
            always the case except in rare situations of wearing metalic clothing.
    """
    return solar_flux * (body_absorptivity / body_emissivity)


def body_solar_flux_from_parts(diff_horiz_solar, dir_normal_solar, altitude,
                               sharp=135, sky_exposure=1, fract_exposed=1,
                               floor_reflectance=0.25, posture='standing'):
    """Estimate the total solar flux on human geometry from solar components.

    Args:
        diff_horiz_solar: Diffuse horizontal solar irradiance in W/m2.
        dir_normal_solar: Direct normal solar irradiance in W/m2.
        alt: The altitude of the sun in degrees [0-90].
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for outdoors in an
            open field.
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include the
            body’s self-shading; only the shading from surroundings.
            Default is 1 for a person standing in an open area.
        floor_reflectance: A number between 0 and 1 the represents the
            reflectance of the floor. Default is for 0.25 which is characteristic
            of outdoor grass or dry bare soil.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
    """
    fract_eff = 0.696 if posture == 'seated' else 0.725
    glob_horiz = diff_horiz_solar + (dir_normal_solar * math.sin(math.radians(altitude)))

    dir_solar = body_dir_from_dir_normal(dir_normal_solar, altitude, sharp,
                                         posture, fract_exposed)
    diff_solar = body_diff_from_diff_horiz(diff_horiz_solar, sky_exposure, fract_eff)
    ref_solar = body_ref_from_glob_horiz(glob_horiz, floor_reflectance,
                                         sky_exposure, fract_eff)
    return dir_solar + diff_solar + ref_solar


def body_solar_flux_from_horiz_parts(diff_horiz_solar, dir_horiz_solar, altitude,
                                     sharp=135, fract_exposed=1,
                                     floor_reflectance=0.25, posture='standing'):
    """Estimate total solar flux on human geometry from horizontal solar components.

    This method is useful for cases when one wants to take the hourly results
    of a spatial radiation study with Radiance and use them to build a map
    of ERF or MRT delta on a person.

    Args:
        diff_horiz_solar: Diffuse horizontal solar irradiance in W/m2.
        dir_horiz_solar: Direct horizontal solar irradiance in W/m2.
        alt: The altitude of the sun in degrees [0-90].
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include the
            body’s self-shading; only the shading from surroundings.
            Default is 1 for a person standing in an open area.
        floor_reflectance: A number between 0 and 1 the represents the
            reflectance of the floor. Default is for 0.25 which is characteristic
            of outdoor grass or dry bare soil.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
    """
    fract_eff = 0.696 if posture == 'seated' else 0.725
    glob_horiz = diff_horiz_solar + dir_horiz_solar

    dir_solar = body_dir_from_dir_horiz(dir_horiz_solar, altitude, sharp,
                                        posture, fract_exposed)
    diff_solar = body_diff_from_diff_horiz(diff_horiz_solar, 1, fract_eff)
    ref_solar = body_ref_from_glob_horiz(glob_horiz, floor_reflectance, 1, fract_eff)
    return dir_solar + diff_solar + ref_solar


def body_diff_from_diff_horiz(diff_horiz_solar, sky_exposure=1, fract_efficiency=0.725):
    """Estimate the diffuse solar flux on human geometry from diffuse horizontal solar.

    Args:
        diff_horiz_solar: Diffuse horizontal solar irradiance in W/m2.
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for outdoors in an
            open field.
        fract_efficiency: A number representing the fraction of the body
            surface exposed to radiation from the environment. This is typically
            either 0.725 for a standing or supine person or 0.696 for a seated
            person. Default is 0.725 for a standing person.
    """
    return 0.5 * sky_exposure * fract_efficiency * diff_horiz_solar


def body_ref_from_glob_horiz(glob_horiz_solar, floor_reflectance=0.25,
                             sky_exposure=1, fract_efficiency=0.725):
    """Estimate floor-reflected solar flux on human geometry from global horizontal solar.

    Args:
        glob_horiz_solar: Global horizontal solar irradiance in W/m2.
        floor_reflectance: A number between 0 and 1 the represents the
            reflectance of the floor. Default is for 0.25 which is characteristic
            of outdoor grass or dry bare soil.
        sky_exposure: A number between 0 and 1 representing the fraction of the
            sky vault in occupant’s view. Default is 1 for outdoors in an
            open field.
        fract_efficiency: A number representing the fraction of the body
            surface exposed to radiation from the environment. This is typically
            either 0.725 for a standing or supine person or 0.696 for a seated
            person. Default is 0.725 for a standing person.
    """
    return 0.5 * sky_exposure * fract_efficiency * glob_horiz_solar * floor_reflectance


def body_dir_from_dir_horiz(dir_horiz_solar, altitude, sharp=135,
                            posture='standing', fract_exposed=1):
    """Estimate the direct solar flux on human geometry from direct horizontal solar.

    Args:
        dir_horiz_solar: Direct horizontal solar irradiance in W/m2.
        altitude: A number between 0 and 90 representing the altitude of the
            sun in degrees.
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include
            the body’s self-shading; only the shading from surroundings.
            Default is 1 for a person in an open area.
    """
    try:
        proj_fac = get_projection_factor(altitude, sharp, posture)
    except KeyError:
        proj_fac = get_projection_factor_simple(altitude, sharp, posture)
    dir_normal_solar = dir_horiz_solar / math.sin(math.radians(altitude))
    return proj_fac * fract_exposed * dir_normal_solar


def body_dir_from_dir_normal(dir_normal_solar, altitude, sharp=135,
                             posture='standing', fract_exposed=1):
    """Estimate the direct solar flux on human geometry from direct horizontal solar.

    Args:
        dir_normal_solar: Direct normal solar irradiance in W/m2.
        altitude: A number between 0 and 90 representing the altitude of the
            sun in degrees.
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
        fract_exposed: A number between 0 and 1 representing the fraction of
            the body exposed to direct sunlight. Note that this does not include
            the body’s self-shading; only the shading from surroundings.
            Default is 1 for a person in an open area.
    """
    try:
        proj_fac = get_projection_factor(altitude, sharp, posture)
    except KeyError:
        proj_fac = get_projection_factor_simple(altitude, sharp, posture)
    return proj_fac * fract_exposed * dir_normal_solar


def sharp_from_solar_and_body_azimuth(solar_azimuth, body_azimuth=0):
    """Calculate solar horizontal angle relative to front of person (SHARP).

    Args:
        solar_azimuth: A number between 0 and 360 representing the solar azimuth
            in degrees (0=North, 90=East, 180=South, 270=West).
        body_azimuth: A number between 0 and 360 representing the direction that
            the human is facing in degrees (0=North, 90=East, 180=South, 270=West).
    """
    angle_diff = abs(solar_azimuth - body_azimuth)
    if angle_diff <= 180:
        return angle_diff
    else:
        return 360 - angle_diff


def get_projection_factor(altitude, sharp=135, posture='standing'):
    """Get the fraction of body surface area exposed to direct sun from solar position.

    This is effectively Ap / Ad in the original Solarcal equations.

    Args:
        altitude: A number between 0 and 90 representing the altitude of the
            sun in degrees.
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). 0 signifies sun that is
            shining directly into the person's face and 180 signifies sun that
            is shining at the person's back. Default is 135, asuming that a person
            typically faces their side or back to the sun to avoid glare.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
    """
    if posture == 'supine':
        altitude, sharp = transpose_altitude_azimuth(altitude, sharp)
        altitude = 1 if altitude == 0 else altitude
        posture = 'standing'
    try:
        return SOLARCAL_SPLINES[posture][int(sharp)][int(math.ceil(altitude) - 1)]
    except IndexError:
        raise ValueError('altitude|azimuth {}|{} is outside of acceptable ranges'.format(
            altitude, sharp))


def get_projection_factor_simple(altitude, sharp=135, posture='standing'):
    """Get the fraction of body surface area exposed to direct sun using a simpler method.

    This is effectively Ap / Ad in the original Solarcal equations.

    This is a more portable version of the get_projection_area() function
    since it does not rely on the large matrix of projection factors
    stored externally in csv files. However, it is less precise since it
    effectively interpolates over the missing parts of the matrix. So this is
    only recommended for cases where such csv files are missing.

    Args:
        altitude: A number between 0 and 90 representing the altitude of the
            sun in degrees.
        sharp: A number between 0 and 180 representing the solar horizontal
            angle relative to front of person (SHARP). Default is 135, asuming
            a person typically faces their side or back to the sun to avoid glare.
        posture: A text string indicating the posture of the body. Letters must
            be lowercase.  Choose from the following: "standing", "seated", "supine".
            Default is "standing".
    """
    if posture == 'supine':
        altitude, sharp = transpose_altitude_azimuth(altitude, sharp)
        posture = 'standing'

    if posture == 'standing':
        ap_table = ((0.254, 0.254, 0.228, 0.187, 0.149, 0.104, 0.059),
                    (0.248, 0.248, 0.225, 0.183, 0.145, 0.102, 0.059),
                    (0.239, 0.239, 0.218, 0.177, 0.138, 0.096, 0.059),
                    (0.225, 0.225, 0.199, 0.165, 0.127, 0.09, 0.059),
                    (0.205, 0.205, 0.182, 0.151, 0.116, 0.083, 0.059),
                    (0.183, 0.183, 0.165, 0.136, 0.109, 0.078, 0.059),
                    (0.167, 0.167, 0.155, 0.131, 0.107, 0.078, 0.059),
                    (0.175, 0.175, 0.161, 0.131, 0.111, 0.081, 0.059),
                    (0.199, 0.199, 0.178, 0.147, 0.12, 0.084, 0.059),
                    (0.22, 0.22, 0.196, 0.16, 0.126, 0.088, 0.059),
                    (0.238, 0.238, 0.21, 0.17, 0.133, 0.091, 0.059),
                    (0.249, 0.249, 0.22, 0.177, 0.138, 0.093, 0.059),
                    (0.252, 0.252, 0.223, 0.178, 0.138, 0.093, 0.059))
    elif posture == 'seated':
        ap_table = ((0.202, 0.226, 0.212, 0.211, 0.182, 0.156, 0.123),
                    (0.203, 0.228, 0.205, 0.2, 0.187, 0.158, 0.123),
                    (0.2, 0.231, 0.207, 0.202, 0.184, 0.155, 0.123),
                    (0.191, 0.227, 0.205, 0.201, 0.175, 0.149, 0.123),
                    (0.177, 0.214, 0.195, 0.192, 0.168, 0.141, 0.123),
                    (0.16, 0.196, 0.182, 0.181, 0.162, 0.134, 0.123),
                    (0.15, 0.181, 0.173, 0.17, 0.153, 0.129, 0.123),
                    (0.163, 0.18, 0.164, 0.158, 0.145, 0.125, 0.123),
                    (0.182, 0.181, 0.156, 0.145, 0.136, 0.122, 0.123),
                    (0.195, 0.181, 0.146, 0.134, 0.128, 0.118, 0.123),
                    (0.207, 0.178, 0.135, 0.121, 0.117, 0.117, 0.123),
                    (0.213, 0.174, 0.125, 0.109, 0.109, 0.116, 0.123),
                    (0.209, 0.167, 0.117, 0.106, 0.106, 0.114, 0.123))
    else:
        raise TypeError('Posture type {} is not recognized.'.format(posture))

    def _find_span(arr, x):
        # For orderd array arr, find the left index of the closest interval x falls in.
        for i in range(len(arr) - 1):
            if x <= arr[i+1] and x >= arr[i]:
                return i
        raise ValueError('altitude/azimuth {} is outside of acceptable ranges'.format(x))

    alt_range = (0, 15, 30, 45, 60, 75, 90)
    az_range = (0, 15, 30, 45, 60, 75, 90, 105, 120, 135, 150, 165, 180)
    alt_i = _find_span(alt_range, altitude)
    az_i = _find_span(az_range, sharp)

    ap11 = ap_table[az_i][alt_i]
    ap12 = ap_table[az_i][alt_i + 1]
    ap21 = ap_table[az_i + 1][alt_i]
    ap22 = ap_table[az_i + 1][alt_i + 1]

    az1 = az_range[az_i]
    az2 = az_range[az_i+1]
    alt1 = alt_range[alt_i]
    alt2 = alt_range[alt_i+1]

    # bilinear interpolation
    ap = ap11 * (az2 - sharp) * (alt2 - altitude)
    ap += ap21 * (sharp - az1) * (alt2 - altitude)
    ap += ap12 * (az2 - sharp) * (altitude - alt1)
    ap += ap22 * (sharp - az1) * (altitude - alt1)
    ap /= (az2 - az1) * (alt2 - alt1)

    return ap


def transpose_altitude_azimuth(altitude, azimuth):
    """Transpose altitude and azimuth.

    This is necessary for getting correct projection factors for a supine posture
    from the standing posture matrix.
    """
    alt_temp = altitude
    altitude = abs(90 - azimuth)
    azimuth = alt_temp
    return altitude, azimuth



# coding=utf-8
"""Parameters for specifying acceptable thermal conditions using the UTCI model."""
from __future__ import division

from ._base import ComfortParameter


class UTCIParameter(ComfortParameter):
    """Parameters of UTCI comfort.

    Properties:
        cold_thresh
        heat_thresh
        extreme_cold_thresh
        very_strong_cold_thresh
        strong_cold_thresh
        moderate_cold_thresh
        moderate_heat_thresh
        strong_heat_thresh
        very_strong_heat_thresh
        extreme_heat_thresh
    """
    _model = 'Universal Thermal Climate Index'

    def __init__(self, cold_thresh=None, heat_thresh=None,
                 extreme_cold_thresh=None, very_strong_cold_thresh=None,
                 strong_cold_thresh=None,
                 moderate_cold_thresh=None, moderate_heat_thresh=None,
                 strong_heat_thresh=None,
                 very_strong_heat_thresh=None, extreme_heat_thresh=None):
        """Initalize UTCI Parameters.

        Args:
            cold_thresh:  UTCI temperature below which conditions
                represent cold stress [C]. Default: 9C.
            heat_thresh:  UTCI temperature above which conditions
                represent heat stress [C]. Default: 26C.
            extreme_cold_thresh:  UTCI temperature below which conditions
                represent extreme cold stress [C]. Default: -40C.
            very_strong_cold_thresh:  UTCI temperature below which conditions
                represent very strong cold stress [C]. Default: -27C.
            strong_cold_thresh:  UTCI temperature below which conditions
                represent strong cold stress [C]. Default: -13C.
            moderate_cold_thresh:  UTCI temperature below which conditions
                represent moderate cold stress [C]. Default: 0C.
            moderate_heat_thresh:  UTCI temperature above which conditions
                represent moderateheat stress [C]. Default: 28C.
            strong_heat_thresh:  UTCI temperature above which conditions
                represent strong heat stress [C]. Default: 32C.
            very_strong_heat_thresh:  UTCI temperature above which conditions
                represent very strong heat stress [C]. Default: 38C.
            extreme_heat_thresh:  UTCI temperature above which conditions
                represent extreme heat stress [C]. Default: 46C.
        """

        self._cold_thresh = cold_thresh if cold_thresh is not None else 9
        self._heat_thresh = heat_thresh if heat_thresh is not None else 26

        self._extreme_cold_thresh = \
            extreme_cold_thresh if extreme_cold_thresh is not None else -40
        self._very_strong_cold_thresh = \
            very_strong_cold_thresh if very_strong_cold_thresh is not None else -27
        self._strong_cold_thresh = \
            strong_cold_thresh if strong_cold_thresh is not None else -13

        self._moderate_cold_thresh = \
            moderate_cold_thresh if moderate_cold_thresh is not None else 0
        self._moderate_heat_thresh = \
            moderate_heat_thresh if moderate_heat_thresh is not None else 28

        self._strong_heat_thresh = \
            strong_heat_thresh if strong_heat_thresh is not None else 32
        self._very_strong_heat_thresh = \
            very_strong_heat_thresh if very_strong_heat_thresh is not None else 38
        self._extreme_heat_thresh = \
            extreme_heat_thresh if extreme_heat_thresh is not None else 46

        assert self._extreme_cold_thresh <= self._very_strong_cold_thresh, \
            'extreme_strong_cold_thresh must be <= very_strong_cold_thresh'
        assert self._very_strong_cold_thresh <= self._strong_cold_thresh, \
            'very_strong_cold_thresh must be <= strong_cold_thresh'
        assert self._strong_cold_thresh <= self._moderate_cold_thresh, \
            'strong_cold_thresh must be <= moderate_cold_thresh'
        assert self._moderate_cold_thresh <= self._cold_thresh, \
            'moderate_cold_thresh must be <= cold_thresh'
        assert self._cold_thresh <= self._heat_thresh, \
            'cold_thresh must be <= heat_thresh'
        assert self._heat_thresh <= self._moderate_heat_thresh, \
            'heat_thresh must be <= moderate_heat_thresh'
        assert self._moderate_heat_thresh <= self._strong_heat_thresh, \
            'moderate_heat_thresh must be <= strong_heat_thresh'
        assert self._strong_heat_thresh <= self._very_strong_heat_thresh, \
            'strong_heat_thresh must be <= very_strong_heat_thresh'
        assert self._very_strong_heat_thresh <= self._extreme_heat_thresh, \
            'very_strong_heat_thresh must be <= extreme_heat_thresh'

    @property
    def cold_thresh(self):
        """UTCI temperature below which conditions represent cold stress [C].

        Default: 9C.
        """
        return self._cold_thresh

    @property
    def heat_thresh(self):
        """UTCI temperature above which conditions represent heat stress [C].

        Default: 26C.
        """
        return self._heat_thresh

    @property
    def extreme_cold_thresh(self):
        """UTCI temperature below which conditions represent extreme cold stress [C].

        Default: -40C.
        """
        return self._extreme_cold_thresh

    @property
    def very_strong_cold_thresh(self):
        """UTCI temperature below which conditions represent very strong cold stress [C].

        Default: -27C.
        """
        return self._very_strong_cold_thresh

    @property
    def strong_cold_thresh(self):
        """UTCI temperature below which conditions represent strong cold stress [C].

        Default: -13C.
        """
        return self._strong_cold_thresh

    @property
    def moderate_cold_thresh(self):
        """UTCI temperature below which conditions represent moderate cold stress [C].

        Default: 0C.
        """
        return self._moderate_cold_thresh

    @property
    def moderate_heat_thresh(self):
        """UTCI temperature above which conditions represent moderate heat stress [C].

        Default: 28C.
        """
        return self._moderate_heat_thresh

    @property
    def strong_heat_thresh(self):
        """UTCI temperature above which conditions represent strong heat stress [C].

        Default: 32C.
        """
        return self._strong_heat_thresh

    @property
    def very_strong_heat_thresh(self):
        """UTCI temperature above which conditions represent very strong heat stress [C].

        Default: 38C.
        """
        return self._very_strong_heat_thresh

    @property
    def extreme_heat_thresh(self):
        """UTCI temperature above which conditions represent extreme heat stress [C].

        Default: 46C.
        """
        return self._extreme_heat_thresh

    def is_comfortable(self, utci):
        """Determine if conditions are comfortable or not.

        Values are one of the following:
            0 = uncomfortable
            1 = comfortable
        """
        return 1 if (utci >= self._cold_thresh and utci <= self._heat_thresh) else 0

    def thermal_condition(self, utci):
        """Determine whether conditions are cold, neutral or hot.

        Values are one of the following:
            -1 = cold
             0 = netural
            +1 = hot
        """
        if utci < self._cold_thresh:
            return -1
        elif utci > self._heat_thresh:
            return 1
        else:
            return 0

    def thermal_condition_five_point(self, utci):
        """Determine the thermal condition on a five-point scale.

        Values are one of the following:
            -2 = strong/extreme cold stress
            -1 = moderate cold stress
             0 = no thermal stress
            +1 = moderate heat stress
            +2 = strong/extreme heat stress
        """
        if utci < self._strong_cold_thresh:
            return -2
        elif utci < self._cold_thresh:
            return -1
        elif utci > self._strong_heat_thresh:
            return 2
        elif utci > self._heat_thresh:
            return 1
        else:
            return 0

    def thermal_condition_seven_point(self, utci):
        """Determine the thermal condition on a seven-point scale.

        Values are one of the following:
            -3 = very strong/extreme cold stress
            -2 = strong cold stress
            -1 = moderate cold stress
             0 = no thermal stress
            +1 = moderate heat stress
            +2 = strong heat stress
            +3 = very strong/extreme heat stress
        """
        if utci < self._very_strong_cold_thresh:
            return -3
        elif utci < self._strong_cold_thresh:
            return -2
        elif utci < self._cold_thresh:
            return -1
        elif utci > self._very_strong_heat_thresh:
            return 3
        elif utci > self._strong_heat_thresh:
            return 2
        elif utci > self._heat_thresh:
            return 1
        else:
            return 0

    def thermal_condition_nine_point(self, utci):
        """Determine the thermal condition on a nine-point scale.

        Values are one of the following:
            -4 = very strong/extreme cold stress
            -3 = strong cold stress
            -2 = moderate cold stress
            -1 = slight cold stress
             0 = no thermal stress
            +1 = slight heat stress
            +2 = moderate heat stress
            +3 = strong heat stress
            +4 = very strong/extreme heat stress
        """
        if utci < self._very_strong_cold_thresh:
            return -4
        elif utci < self._strong_cold_thresh:
            return -3
        elif utci < self._moderate_cold_thresh:
            return -2
        elif utci < self._cold_thresh:
            return -1
        elif utci > self._very_strong_heat_thresh:
            return 4
        elif utci > self._strong_heat_thresh:
            return 3
        elif utci > self._moderate_heat_thresh:
            return 2
        elif utci > self._heat_thresh:
            return 1
        else:
            return 0

    def thermal_condition_eleven_point(self, utci):
        """Determine the thermal condition on an eleven-point scale.

        Values are one of the following:
            -5 = extreme cold stress
            -4 = very strong cold stress
            -3 = strong cold stress
            -2 = moderate cold stress
            -1 = slight cold stress
             0 = no thermal stress
            +1 = slight heat stress
            +2 = moderate heat stress
            +3 = strong heat stress
            +4 = very strong heat stress
            +5 = extreme heat stress
        """
        if utci < self._extreme_cold_thresh:
            return -5
        elif utci < self._very_strong_cold_thresh:
            return -4
        elif utci < self._strong_cold_thresh:
            return -3
        elif utci < self._moderate_cold_thresh:
            return -2
        elif utci < self._cold_thresh:
            return -1
        elif utci > self._extreme_heat_thresh:
            return 5
        elif utci > self._very_strong_heat_thresh:
            return 4
        elif utci > self._strong_heat_thresh:
            return 3
        elif utci > self._moderate_heat_thresh:
            return 2
        elif utci > self._heat_thresh:
            return 1
        else:
            return 0

    def original_utci_category(self, utci):
        """Determine the category according to the original UTCI assessment scale.

        Glossary of Terms for Thermal Physiology (2003).
        Journal of Thermal Biology 28, 75-106

        Values are one of the following:
            0 = extreme cold stress
            1 = very strong cold stress
            2 = strong cold stress
            3 = moderate cold stress
            4 = slight cold stress
            5 = no thermal stress
            6 = moderate heat stress
            7 = strong heat stress
            8 = strong heat stress
            9 = extreme heat stress
        """
        if utci < self._extreme_cold_thresh:
            return 0
        elif utci < self._very_strong_cold_thresh:
            return 1
        elif utci < self._strong_cold_thresh:
            return 2
        elif utci < self._moderate_cold_thresh:
            return 3
        elif utci < self._cold_thresh:
            return 4
        elif utci > self._extreme_heat_thresh:
            return 9
        elif utci > self._very_strong_heat_thresh:
            return 8
        elif utci > self._strong_heat_thresh:
            return 7
        elif utci > self._heat_thresh:
            return 6
        else:
            return 5

    def duplicate(self):
        """Duplicate these comfort parameters."""
        return UTCIParameter(self.cold_thresh, self.heat_thresh,
                             self.extreme_cold_thresh, self.very_strong_cold_thresh,
                             self.strong_cold_thresh,
                             self.moderate_cold_thresh, self.moderate_heat_thresh,
                             self.strong_heat_thresh,
                             self.very_strong_heat_thresh, self.extreme_heat_thresh)

    def __repr__(self):
        """UTCI comfort parameters representation."""
        return "UTCI Comfort Parameters\n Cold Threshold: {}C"\
            "\n Heat Threshold: {}C".format(
                self._cold_thresh, self._heat_thresh)



# coding=utf-8
"""Object for calculating UTCI comfort from DataCollections."""
from __future__ import division

from ..utci import universal_thermal_climate_index
from ..parameter.utci import UTCIParameter
from .base import ComfortCollection

from ladybug._datacollectionbase import BaseCollection

from ladybug.datatype.temperature import Temperature, MeanRadiantTemperature, \
    AirTemperature, UniversalThermalClimateIndex
from ladybug.datatype.fraction import RelativeHumidity
from ladybug.datatype.speed import Speed, WindSpeed
from ladybug.datatype.thermalcondition import ThermalComfort, ThermalCondition, \
    ThermalConditionFivePoint, ThermalConditionSevenPoint, \
    ThermalConditionNinePoint, ThermalConditionElevenPoint, UTCICategory


class UTCI(ComfortCollection):
    """UTCI comfort DataCollection object.

    Properties:
        air_temperature
        rad_temperature
        air_speed
        rel_humidity
        comfort_parameter
        universal_thermal_climate_index
        is_comfortable
        thermal_condition
        thermal_condition_five_point
        thermal_condition_seven_point
        thermal_condition_nine_point
        thermal_condition_eleven_point
        original_utci_category
        percent_comfortable
        percent_uncomfortable
        percent_neutral
        percent_hot
        percent_cold
        percent_extreme_cold_stress
        percent_very_strong_cold_stress
        percent_strong_cold_stress
        percent_moderate_cold_stress
        percent_slight_cold_stress
        percent_slight_heat_stress
        percent_moderate_heat_stress
        percent_strong_heat_stress
        percent_very_strong_heat_stress
        percent_extreme_heat_stress
    """
    _model = 'Universal Thermal Climate Index'

    def __init__(self, air_temperature, rel_humidity, rad_temperature=None,
                 wind_speed=None, comfort_parameter=None):
        """Initialize a UTCI comfort object from DataCollections of UTCI inputs.

        Args:
            air_temperature: Data Collection of air temperature values in Celcius.
            rel_humidity: Data Collection of relative humidity values in % or a
                single relative humdity value to be used for the whole analysis.
            rad_temperature: Data Collection of mean radiant temperature (MRT)
                values in degrees Celcius or a single MRT value to be used for the whole
                analysis. If None, this will be the same as the air_temperature.
            wind_speed: Data Collection of meteorological wind speed values in m/s
                (measured 10 m above the ground) or a single wind speed value to be
                used for the whole analysis. If None, this will default to a very
                low wind speed of 0.1 m/s.
            comfort_parameter: Optional UTCIParameter object to specify parameters under
                which conditions are considered acceptable. If None, default will
                assume comfort thresholds consistent with those used by meterologists
                to categorize outdoor conditions.
        """
        # set up the object using air temperature as a base
        self._check_datacoll(air_temperature, Temperature, 'C', 'air_temperature')
        self._input_collections = [air_temperature]
        self._calc_length = len(air_temperature.values)
        self._base_collection = air_temperature

        # check required inputs
        self._air_temperature = air_temperature.values
        self._rel_humidity = self._check_input(
            rel_humidity, RelativeHumidity, '%', 'rel_humidity')

        # check inputs with defaults
        if rad_temperature is not None:
            self._rad_temperature = self._check_input(
                rad_temperature, Temperature, 'C', 'rad_temperature')
        else:
            self._rad_temperature = self._air_temperature

        if wind_speed is not None:
            self._wind_speed = self._check_input(
                wind_speed, Speed, 'm/s', 'air_speed')
        else:
            self._wind_speed = [0.1] * self.calc_length

        # check that all input data collections are aligned.
        BaseCollection.are_collections_aligned(self._input_collections)

        # check comfort parameters
        if comfort_parameter is None:
            self._comfort_par = UTCIParameter()
        else:
            assert isinstance(comfort_parameter, UTCIParameter), 'comfort_parameter '\
                'must be a UTCIParameter object. Got {}'.format(type(comfort_parameter))
            self._comfort_par = comfort_parameter

        # compute UTCI
        self._calculate_utci()

    def _calculate_utci(self):
        """Compute UTCI for each step of the Data Collection."""
        self._utci = []
        self._thermal_category = []
        for ta, tr, vel, rh in \
            zip(self._air_temperature, self._rad_temperature,
                self._wind_speed, self._rel_humidity):
            result = universal_thermal_climate_index(ta, tr, vel, rh)
            self._utci.append(result)
            self._thermal_category.append(
                self._comfort_par.thermal_condition_eleven_point(result))

    @property
    def air_temperature(self):
        """Data Collection of air temperature values in degrees C."""
        return self._get_coll('_air_temperature_coll', self._air_temperature,
                              AirTemperature, 'C')

    @property
    def rad_temperature(self):
        """Data Collection of mean radiant temperature (MRT) values in degrees C."""
        return self._get_coll('_rad_temperature_coll', self._rad_temperature,
                              MeanRadiantTemperature, 'C')

    @property
    def wind_speed(self):
        """Data Collection of air speed values in m/s."""
        return self._get_coll('_wind_speed_coll', self._wind_speed,
                              WindSpeed, 'm/s')

    @property
    def rel_humidity(self):
        """Data Collection of relative humidity values in %."""
        return self._get_coll('_rel_humidity_coll', self._rel_humidity,
                              RelativeHumidity, '%')

    @property
    def comfort_parameter(self):
        """UTCI comfort parameters that are assigned to this object."""
        return self._comfort_par

    @property
    def universal_thermal_climate_index(self):
        """A Data Collection of Universal Thermal Climate Index (UTCI) in C."""
        return self._get_coll('_utci_coll', self._utci,
                              UniversalThermalClimateIndex, 'C')

    @property
    def is_comfortable(self):
        """Data Collection of integers noting whether the input conditions are
        acceptable according to the assigned comfort_parameter.

        Values are one of the following:
            0 = uncomfortable
            1 = comfortable
        """
        return self._get_coll('_is_comfortable_coll', self._comf_val_funct,
                              ThermalComfort, 'condition')

    @property
    def thermal_condition(self):
        """Data Collection of integers noting the thermal status of a subject
        according to the assigned comfort_parameter.

        Values are one of the following:
            -1 = cold
             0 = netural
            +1 = hot
        """
        return self._get_coll('_thermal_condition_coll', self._condit_val_funct,
                              ThermalCondition, 'condition')

    @property
    def thermal_condition_five_point(self):
        """Data Collection of integers noting the thermal status on a five-point scale.

        Values are one of the following:
            -2 = strong/extreme cold stress
            -1 = moderate cold stress
             0 = no thermal stress
            +1 = moderate heat stress
            +2 = strong/extreme heat stress
        """
        return self._get_coll('_five_point_coll', self._five_pt_funct,
                              ThermalConditionFivePoint, 'condition')

    @property
    def thermal_condition_seven_point(self):
        """Data Collection of integers noting the thermal status on a seven-point scale.

        Values are one of the following:
            -3 = very strong/extreme cold stress
            -2 = strong cold stress
            -1 = moderate cold stress
             0 = no thermal stress
            +1 = moderate heat stress
            +2 = strong heat stress
            +3 = very strong/extreme heat stress
        """
        return self._get_coll('_seven_point_coll', self._seven_pt_funct,
                              ThermalConditionSevenPoint, 'condition')

    @property
    def thermal_condition_nine_point(self):
        """Data Collection of integers noting the thermal status on a nine-point scale.

        Values are one of the following:
            -4 = very strong/extreme cold stress
            -3 = strong cold stress
            -2 = moderate cold stress
            -1 = slight cold stress
             0 = no thermal stress
            +1 = slight heat stress
            +2 = moderate heat stress
            +3 = strong heat stress
            +4 = very strong/extreme heat stress
        """
        return self._get_coll('_nine_point_coll', self._nine_pt_funct,
                              ThermalConditionNinePoint, 'condition')

    @property
    def thermal_condition_eleven_point(self):
        """Data Collection of integers noting the thermal status on an eleven-point scale.

        Values are one of the following:
            -5 = extreme cold stress
            -4 = very strong cold stress
            -3 = strong cold stress
            -2 = moderate cold stress
            -1 = slight cold stress
             0 = no thermal stress
            +1 = slight heat stress
            +2 = moderate heat stress
            +3 = strong heat stress
            +4 = very strong heat stress
            +5 = extreme heat stress
        """
        return self._get_coll('_eleven_point_coll', self._thermal_category,
                              ThermalConditionElevenPoint, 'condition')

    @property
    def original_utci_category(self):
        """Data Collection of integers noting the original UTCI assessment scale.

        Glossary of Terms for Thermal Physiology (2003).
        Journal of Thermal Biology 28, 75-106

        Values are one of the following:
            0 = extreme cold stress
            1 = very strong cold stress
            2 = strong cold stress
            3 = moderate cold stress
            4 = slight cold stress
            5 = no thermal stress
            6 = moderate heat stress
            7 = strong heat stress
            8 = strong heat stress
            9 = extreme heat stress
        """
        return self._get_coll('_original_category_coll', self._original_category_funct,
                              UTCICategory, 'condition')

    @property
    def percent_comfortable(self):
        """The percent of time comfortabe given by the assigned comfort_parameter."""
        _vals = [1 for t in self._thermal_category if t == 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_uncomfortable(self):
        """The percent of time uncomfortabe given by the assigned comfort_parameter."""
        return 100 - self.percent_comfortable

    @property
    def percent_neutral(self):
        """The percent of time that the thermal_condiiton is neutral."""
        return self.percent_comfortable

    @property
    def percent_cold(self):
        """The percent of time that the thermal_condiiton is cold."""
        _vals = [1 for x in self._thermal_category if x < 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_hot(self):
        """The percent of time that the thermal_condiiton is hot."""
        _vals = [1 for x in self._thermal_category if x > 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_slight_cold_stress(self):
        """The percent of time that conditions have slight cold stress."""
        _vals = [1 for x in self._thermal_category if x == -1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_moderate_cold_stress(self):
        """The percent of time that conditions have moderate cold stress."""
        _vals = [1 for x in self._thermal_category if x == -2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_strong_cold_stress(self):
        """The percent of time that conditions have strong cold stress."""
        _vals = [1 for x in self._thermal_category if x == -3]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_very_strong_cold_stress(self):
        """The percent of time that conditions have very strong cold stress."""
        _vals = [1 for x in self._thermal_category if x == -4]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_extreme_cold_stress(self):
        """The percent of time that conditions have very strong cold stress."""
        _vals = [1 for x in self._thermal_category if x == -5]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_slight_heat_stress(self):
        """The percent of time that conditions have slight heat stress."""
        _vals = [1 for x in self._thermal_category if x == 1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_moderate_heat_stress(self):
        """The percent of time that conditions have moderate heat stress."""
        _vals = [1 for x in self._thermal_category if x == 2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_strong_heat_stress(self):
        """The percent of time that conditions have strong heat stress."""
        _vals = [1 for x in self._thermal_category if x == 3]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_very_strong_heat_stress(self):
        """The percent of time that conditions have very strong heat stress."""
        _vals = [1 for x in self._thermal_category if x == 4]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_extreme_heat_stress(self):
        """The percent of time that conditions have very strong heat stress."""
        _vals = [1 for x in self._thermal_category if x == 5]
        return (sum(_vals) / self._calc_length) * 100

    def _comf_val_funct(self):
        return [self._comfort_par.is_comfortable(t) for t in self._utci]

    def _condit_val_funct(self):
        return [self._comfort_par.thermal_condition(t) for t in self._utci]

    def _five_pt_funct(self):
        return [self._comfort_par.thermal_condition_five_point(t) for t in self._utci]

    def _seven_pt_funct(self):
        return [self._comfort_par.thermal_condition_seven_point(t) for t in self._utci]

    def _nine_pt_funct(self):
        return [self._comfort_par.thermal_condition_nine_point(t) for t in self._utci]

    def _original_category_funct(self):
        return [self._comfort_par.original_utci_category(t) for t in self._utci]



# coding utf-8
import pytest

from ladybug_comfort.collection.utci import UTCI
from ladybug_comfort.collection.pmv import PMV

from ladybug.datacollection import HourlyContinuousCollection
from ladybug.epw import EPW


def test_get_universal_thermal_climate_index():
    """Test the get_universal_thermal_climate_index method."""
    calc_length = 8760
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    utci_obj = epw.get_universal_thermal_climate_index(False, False)

    assert isinstance(utci_obj, UTCI)
    assert isinstance(utci_obj.air_temperature, HourlyContinuousCollection)
    assert len(utci_obj.air_temperature.values) == calc_length
    assert utci_obj.air_temperature[0] == -6.1
    assert isinstance(utci_obj.rel_humidity, HourlyContinuousCollection)
    assert len(utci_obj.rel_humidity.values) == calc_length
    assert utci_obj.rel_humidity[0] == 81

    assert isinstance(utci_obj.universal_thermal_climate_index, HourlyContinuousCollection)
    assert len(utci_obj.universal_thermal_climate_index.values) == calc_length
    assert utci_obj.universal_thermal_climate_index[0] == pytest.approx(-5.367017, rel=1e-3)
    assert isinstance(utci_obj.thermal_condition_eleven_point, HourlyContinuousCollection)
    assert len(utci_obj.thermal_condition_eleven_point.values) == calc_length
    assert utci_obj.thermal_condition_eleven_point[0] == -2

    assert utci_obj.percent_neutral == pytest.approx(47.488584, rel=1e-3)
    assert utci_obj.percent_hot == pytest.approx(9.38356164, rel=1e-3)
    assert utci_obj.percent_cold == pytest.approx(43.1278538, rel=1e-3)


def test_get_universal_thermal_climate_index_with_wind():
    """Test the get_universal_thermal_climate_index method with wind."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    utci_obj = epw.get_universal_thermal_climate_index(True, False)

    assert utci_obj.percent_neutral == pytest.approx(35.6849315, rel=1e-3)
    assert utci_obj.percent_hot == pytest.approx(3.3447488, rel=1e-3)
    assert utci_obj.percent_cold == pytest.approx(60.970319, rel=1e-3)


def test_get_universal_thermal_climate_index_with_sun():
    """Test the get_universal_thermal_climate_index method with sun."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    utci_obj = epw.get_universal_thermal_climate_index(False, True)

    assert utci_obj.percent_neutral == pytest.approx(40.730593, rel=1e-3)
    assert utci_obj.percent_hot == pytest.approx(20.4223744, rel=1e-3)
    assert utci_obj.percent_cold == pytest.approx(38.8470319, rel=1e-3)


def test_get_universal_thermal_climate_index_with_sun_and_wind():
    """Test the get_universal_thermal_climate_index method with wind and sun."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    utci_obj = epw.get_universal_thermal_climate_index(True, True)

    assert utci_obj.percent_neutral == pytest.approx(31.4269406, rel=1e-3)
    assert utci_obj.percent_hot == pytest.approx(11.4611872, rel=1e-3)
    assert utci_obj.percent_cold == pytest.approx(57.111872, rel=1e-3)


def test_get_standard_effective_temperature():
    """Test the get_standard_effective_temperature method."""
    calc_length = 8760
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    set_obj = epw.get_standard_effective_temperature(False, False,
                                                     met_rate=2.4, clo_value=1.0)

    assert isinstance(set_obj, PMV)
    assert isinstance(set_obj.air_temperature, HourlyContinuousCollection)
    assert len(set_obj.air_temperature.values) == calc_length
    assert set_obj.air_temperature[0] == -6.1
    assert isinstance(set_obj.rel_humidity, HourlyContinuousCollection)
    assert len(set_obj.rel_humidity.values) == calc_length
    assert set_obj.rel_humidity[0] == 81

    assert isinstance(set_obj.standard_effective_temperature, HourlyContinuousCollection)
    assert len(set_obj.standard_effective_temperature.values) == calc_length
    assert set_obj.standard_effective_temperature[0] == pytest.approx(12.76, rel=1e-2)
    assert isinstance(set_obj.thermal_condition, HourlyContinuousCollection)
    assert len(set_obj.thermal_condition.values) == calc_length
    assert set_obj.thermal_condition[0] == -1

    assert set_obj.percent_neutral == pytest.approx(18.961187, rel=1e-3)
    assert set_obj.percent_hot == pytest.approx(38.6415525, rel=1e-3)
    assert set_obj.percent_cold == pytest.approx(42.39726, rel=1e-3)


def test_get_standard_effective_temperature_with_wind():
    """Test the get_standard_effective_temperature method with wind."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    set_obj = epw.get_standard_effective_temperature(True, False,
                                                     met_rate=2.4, clo_value=1.0)

    assert set_obj.percent_neutral == pytest.approx(19.82, rel=1e-2)
    assert set_obj.percent_hot == pytest.approx(15.56, rel=1e-2)
    assert set_obj.percent_cold == pytest.approx(64.61, rel=1e-2)


def test_get_standard_effective_temperature_with_sun():
    """Test the get_standard_effective_temperature method with sun."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    set_obj = epw.get_standard_effective_temperature(False, True,
                                                     met_rate=2.4, clo_value=1.0)

    assert set_obj.percent_neutral == pytest.approx(17.043378, rel=1e-3)
    assert set_obj.percent_hot == pytest.approx(44.56621, rel=1e-3)
    assert set_obj.percent_cold == pytest.approx(38.39041, rel=1e-3)


def test_get_standard_effective_temperature_with_sun_and_wind():
    """Test the get_standard_effective_temperature method with sun and wind."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    set_obj = epw.get_standard_effective_temperature(True, True,
                                                     met_rate=2.4, clo_value=1.0)

    assert set_obj.percent_neutral == pytest.approx(17.95, rel=1e-2)
    assert set_obj.percent_hot == pytest.approx(18.82, rel=1e-2)
    assert set_obj.percent_cold == pytest.approx(63.23, rel=1e-2)



# coding=utf-8
"""Object for calculating PMV comfort from DataCollections."""
from __future__ import division

from ..pmv import predicted_mean_vote
from ..parameter.pmv import PMVParameter
from .base import ComfortCollection

from ladybug._datacollectionbase import BaseCollection
from ladybug.psychrometrics import humid_ratio_from_db_rh

from ladybug.datatype.temperature import Temperature, MeanRadiantTemperature, \
    StandardEffectiveTemperature, AirTemperature
from ladybug.datatype.fraction import RelativeHumidity, HumidityRatio, \
    PercentagePeopleDissatisfied
from ladybug.datatype.speed import Speed, AirSpeed
from ladybug.datatype.energyflux import MetabolicRate
from ladybug.datatype.rvalue import ClothingInsulation
from ladybug.datatype.thermalcondition import PredictedMeanVote, \
    ThermalComfort, ThermalCondition, DiscomfortReason
from ladybug.datatype.temperaturedelta import AirTemperatureDelta
from ladybug.datatype.power import Power

try:
    from itertools import izip as zip  # python 2
except ImportError:
    pass


class PMV(ComfortCollection):
    """PMV comfort DataCollection object.

    Properties:
        air_temperature
        rad_temperature
        air_speed
        rel_humidity
        met_rate
        clo_value
        external_work
        comfort_parameter
        predicted_mean_vote
        percentage_people_dissatisfied
        standard_effective_temperature
        is_comfortable
        thermal_condition
        discomfort_reason
        percent_comfortable
        percent_uncomfortable
        percent_neutral
        percent_hot
        percent_cold
        percent_dry
        percent_humid
        humidity_ratio
        adjusted_air_temperature
        cooling_effect
        heat_loss_conduction
        heat_loss_sweating
        heat_loss_latent_respiration
        heat_loss_dry_respiration
        heat_loss_radiation
        heat_loss_convection
    """
    _model = 'Predicted Mean Vote'

    def __init__(self, air_temperature, rel_humidity,
                 rad_temperature=None, air_speed=None,
                 met_rate=None, clo_value=None, external_work=None,
                 comfort_parameter=None):
        """Initialize a PMV comfort object from DataCollections of PMV inputs.

        Args:
            air_temperature: Data Collection of air temperature values in Celcius.
            rel_humidity: Data Collection of relative humidity values in % or a
                single relative humdity value to be used for the whole analysis.
            rad_temperature: Data Collection of mean radiant temperature (MRT)
                values in degrees Celcius or a single MRT value to be used for the whole
                analysis. If None, this will be the same as the air_temperature.
            air_speed: Data Collection of air speed values in m/s or a single
                air_speed value to be used for the whole analysis. If None, this
                will default to 0.1 m/s.
            met_rate: Data Collection of metabolic rate in met or a single
                metabolic rate value to be used for the whole analysis. If None,
                default is set to 1.1 met (for seated, typing).
            clo_value: Data Collection of clothing values rate in clo or a single
                clothing value to be used for the whole analysis. If None, default is
                set to 0.7 clo (for long sleeve shirt and pants).
            external_work: Data Collection of external work in met or a single
                external work value to be used for the whole analysis. If None,
                default is set to 0 met.
            comfort_parameter: Optional PMVParameter object to specify parameters under
                which conditions are considered acceptable. If None, default will
                assume a PPD threshold of 10%, no absolute humidity constraints
                and a still air threshold of 0.1 m/s.
        """
        # set up the object using air temperature as a base
        self._check_datacoll(air_temperature, Temperature, 'C', 'air_temperature')
        self._input_collections = [air_temperature]
        self._calc_length = len(air_temperature.values)
        self._base_collection = air_temperature

        # check and set required inputs
        self._air_temperature = air_temperature.values
        self._rel_humidity = self._check_input(
            rel_humidity, RelativeHumidity, '%', 'rel_humidity')

        # check parameters with defaults
        if rad_temperature is not None:
            self._rad_temperature = self._check_input(
                rad_temperature, Temperature, 'C', 'rad_temperature')
        else:
            self._rad_temperature = self._air_temperature

        if air_speed is not None:
            self._air_speed = self._check_input(
                air_speed, Speed, 'm/s', 'air_speed')
        else:
            self._air_speed = [0.1] * self.calc_length

        if met_rate is not None:
            self._met_rate = self._check_input(
                met_rate, MetabolicRate, 'met', 'met_rate')
        else:
            self._met_rate = [1.1] * self.calc_length

        if clo_value is not None:
            self._clo_value = self._check_input(
                clo_value, ClothingInsulation, 'clo', 'clo_value')
        else:
            self._clo_value = [0.7] * self.calc_length

        if external_work is not None:
            self._external_work = self._check_input(
                external_work, MetabolicRate, 'met', 'external_work')
        else:
            self._external_work = [0.] * self.calc_length

        # check that all input data collections are aligned.
        BaseCollection.are_collections_aligned(self._input_collections)

        # check comfort parameters
        if comfort_parameter is None:
            self._comfort_par = PMVParameter()
        else:
            assert isinstance(comfort_parameter, PMVParameter), 'comfort_parameter '\
                'must be a PMVParameter object. Got {}'.format(type(comfort_parameter))
            self._comfort_par = comfort_parameter

        # value to track whether humidity ratio has been computed
        self._hr_calculated = False
        self._hr_comfort_required = True
        if self.comfort_parameter.humid_ratio_lower == 0 and \
                self.comfort_parameter.humid_ratio_upper == 1:
            self._hr_comfort_required = False

        # calculate PMV
        self._calculate_pmv()

    def _calculate_humidity_ratio(self):
        """Compute the humidity ratio at each step of the Data Collection."""
        self._humidity_ratio = [humid_ratio_from_db_rh(db, rh) for db, rh in zip(
            self._air_temperature, self._rel_humidity)]
        self._hr_calculated = True

    def _calculate_pmv(self):
        """Compute PMV for each step of the Data Collection."""
        if self._hr_comfort_required is True:
            self._calculate_humidity_ratio()

        # empty properties to be caulculated
        self._pmv = []
        self._ppd = []
        self._set = []
        self._is_comfortable = []
        self._thermal_condition = []
        self._discomfort_reason = []
        self._ta_adj = []
        self._cooling_effect = []
        self._heat_loss_conduction = []
        self._heat_loss_sweating = []
        self._heat_loss_latent_respiration = []
        self._heat_loss_dry_respiration = []
        self._heat_loss_radiation = []
        self._heat_loss_convection = []

        # perform the PMV calculation
        for ta, tr, vel, rh, met, clo, wme, i in \
            zip(self._air_temperature, self._rad_temperature,
                self._air_speed, self._rel_humidity,
                self._met_rate, self._clo_value,
                self._external_work, range(self._calc_length)):
            result = predicted_mean_vote(ta, tr, vel, rh, met, clo, wme,
                                         self._comfort_par.still_air_threshold)
            self._pmv.append(result['pmv'])
            self._ppd.append(result['ppd'])
            self._set.append(result['set'])
            self._ta_adj.append(result['ta_adj'])
            self._cooling_effect.append(result['ce'])
            self._heat_loss_conduction.append(result['heat_loss']['cond'])
            self._heat_loss_sweating.append(result['heat_loss']['sweat'])
            self._heat_loss_latent_respiration.append(result['heat_loss']['res_l'])
            self._heat_loss_dry_respiration.append(result['heat_loss']['res_s'])
            self._heat_loss_radiation.append(result['heat_loss']['rad'])
            self._heat_loss_convection.append(result['heat_loss']['conv'])

            # determine whether conditions are acceptable
            condit = self._comfort_par.thermal_condition(
                result['pmv'], result['ppd'])
            if self._hr_comfort_required is True:
                hr = self._humidity_ratio[i]
                comf = self._comfort_par.is_comfortable(result['ppd'], hr)
                reason = self._comfort_par.discomfort_reason(
                    result['pmv'], result['ppd'], hr)
            else:
                comf = self._comfort_par.is_comfortable(result['ppd'])
                reason = self._comfort_par.discomfort_reason(
                    result['pmv'], result['ppd'])
            self._is_comfortable.append(comf)
            self._thermal_condition.append(condit)
            self._discomfort_reason.append(reason)

    @property
    def air_temperature(self):
        """Data Collection of air temperature values in degrees C."""
        return self._get_coll('_air_temperature_coll', self._air_temperature,
                              AirTemperature, 'C')

    @property
    def rad_temperature(self):
        """Data Collection of mean radiant temperature (MRT) values in degrees C."""
        return self._get_coll('_rad_temperature_coll', self._rad_temperature,
                              MeanRadiantTemperature, 'C')

    @property
    def air_speed(self):
        """Data Collection of air speed values in m/s."""
        return self._get_coll('_air_speed_coll', self._air_speed,
                              AirSpeed, 'm/s')

    @property
    def rel_humidity(self):
        """Data Collection of relative humidity values in %."""
        return self._get_coll('_rel_humidity_coll', self._rel_humidity,
                              RelativeHumidity, '%')

    @property
    def met_rate(self):
        """Data Collection of metabolic rate in met.

        1 met = Metabolic rate of a resting seated person
        1.2 met = Metabolic rate of a standing person
        2 met = Metabolic rate of a wlaking person
        If left blank, default is set to 1.1 met (for seated, typing).
        """
        return self._get_coll('_met_rate_coll', self._met_rate,
                              MetabolicRate, 'met')

    @property
    def clo_value(self):
        """Data Collection of clothing level of the human subject in clo.

        1 clo = Three-piece suit
        0.5 clo = Shorts + T-shirt
        0 clo = No clothing
        If left blank, default is set to 0.85 clo.
        """
        return self._get_coll('_clo_value_coll', self._clo_value,
                              ClothingInsulation, 'clo')

    @property
    def external_work(self):
        """Data Collection of the work done by the human subject in met."""
        return self._get_coll('_external_work_coll', self._external_work,
                              MetabolicRate, 'met')

    @property
    def comfort_parameter(self):
        """PMV comfort parameters that are assigned to this object."""
        return self._comfort_par.duplicate()  # duplicate since ppd_thresh is setable

    @property
    def predicted_mean_vote(self):
        """Data Collection of predicted mean vote (PMV) for the input conditions.

        PMV is a seven-point scale from cold (-3) to hot (+3) that was used in comfort
        surveys of P.O. Fanger.
        Each interger value of the scale indicates the following:
            -3 = Cold
            -2 = Cool
            -1 = Slightly Cool
             0 = Neutral
            +1 = Slightly Warm
            +2 = Warm
            +3 = Hot
        """
        return self._get_coll('_pmv_coll', self._pmv, PredictedMeanVote, 'PMV')

    @property
    def percentage_people_dissatisfied(self):
        """Data Collection of percentage of people dissatisfied (PPD) for the input conditions.

        Specifically, this is defined by the percent of people who would have
        a PMV beyond acceptable thresholds (typically <-0.5 and >+0.5).
        Note that, with the PMV model, the best possible PPD achievable is 5%
        and most standards aim to have a PPD below 10%.
        """
        return self._get_coll('_ppd_coll', self._ppd, PercentagePeopleDissatisfied, '%')

    @property
    def standard_effective_temperature(self):
        """Data Collection of standard effective temperature (SET) for the input conditions.

        These temperatures describe what the given input conditions "feel like" in
        relation to a standard environment of 50% relative humidity, <0.1 m/s average
        air speed, and mean radiant temperature equal to average air temperature, in
        which the total heat loss from the skin of an imaginary occupant with an activity
        level of 1.0 met and a clothing level of 0.6 clo is the same as that from a
        person in the actual environment.
        """
        return self._get_coll('_set_coll', self._set,
                              StandardEffectiveTemperature, 'C')

    @property
    def is_comfortable(self):
        """Data Collection of integers noting whether the input conditions are
        acceptable according to the assigned comfort_parameter.

        Values are one of the following:
            0 = uncomfortable
            1 = comfortable
        """
        return self._get_coll('_is_comfortable_coll', self._is_comfortable,
                              ThermalComfort, 'condition')

    @property
    def thermal_condition(self):
        """Data Collection of integers noting the thermal status of a subject
        according to the assigned comfort_parameter.

        Values are one of the following:
            -1 = cold
             0 = netural
            +1 = hot
        """
        return self._get_coll('_thermal_condition_coll', self._thermal_condition,
                              ThermalCondition, 'condition')

    @property
    def discomfort_reason(self):
        """Data Collection of integers noting the reason for discomfort
        according to the assigned comfort_parameter.

        Values are one of the following:
            -2 = too dry
            -1 = too cold
             0 = comfortable
            +1 = too hot
            +2 = too humid
        """
        return self._get_coll('_discomfort_reason_coll', self._discomfort_reason,
                              DiscomfortReason, 'condition')

    @property
    def percent_comfortable(self):
        """The percent of time comfortabe given by the assigned comfort_parameter."""
        return (sum(self._is_comfortable) / self._calc_length) * 100

    @property
    def percent_uncomfortable(self):
        """The percent of time uncomfortabe given by the assigned comfort_parameter."""
        return 100 - self.percent_comfortable

    @property
    def percent_neutral(self):
        """The percent of time that the thermal_condiiton is neutral."""
        _vals = [1 for x in self._thermal_condition if x == 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_cold(self):
        """The percent of time that the thermal_condiiton is cold."""
        _vals = [1 for x in self._thermal_condition if x == -1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_hot(self):
        """The percent of time that the thermal_condiiton is hot."""
        _vals = [1 for x in self._thermal_condition if x == 1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_dry(self):
        """The percent of time that the thermal_condiiton neutral but it is too dry."""
        _vals = [1 for x in self._discomfort_reason if x == -2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_humid(self):
        """The percent of time that the thermal_condiiton neutral but it is too humid."""
        _vals = [1 for x in self._discomfort_reason if x == 2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def humidity_ratio(self):
        """Data Collection of humidity ratio for the dry bulb and relative humidity."""
        if self._hr_calculated is False:
            self._calculate_humidity_ratio()
        return self._get_coll('_humidity_ratio_coll', self._humidity_ratio,
                              HumidityRatio, 'fraction')

    @property
    def adjusted_air_temperature(self):
        """Data Collection of air temperatures that have been adjusted by the SET model
        to account for the effect of air speed [C].
        """
        return self._get_coll('_ta_adj_coll', self._ta_adj,
                              AirTemperature('Adjusted Air Temperature'), 'C')

    @property
    def cooling_effect(self):
        """Data Collection of the cooling effect of the air speed in degrees Celcius.

        This is the difference between the air temperature and the
        adjusted air temperature [C].
        """
        return self._get_coll('_cooling_effect_coll', self._cooling_effect,
                              AirTemperatureDelta('Cooling Effect'), 'C')

    @property
    def heat_loss_conduction(self):
        """Data Collection of heat loss by conduction in [W]."""
        return self._get_coll('_hl_conduction_coll', self._heat_loss_conduction,
                              Power('Heat Loss From Conduction'), 'W')

    @property
    def heat_loss_sweating(self):
        """Data Collection of heat loss by sweating in [W]."""
        return self._get_coll('_hl_sweating', self._heat_loss_sweating,
                              Power('Heat Loss From Sweating'), 'W')

    @property
    def heat_loss_latent_respiration(self):
        """Data Collection of heat loss by latent respiration in [W]."""
        return self._get_coll(
            '_hl_latent_respiration_coll', self._heat_loss_latent_respiration,
            Power('Heat Loss From Latent Respiration'), 'W')

    @property
    def heat_loss_dry_respiration(self):
        """Data Collection of heat loss by dry respiration in [W]."""
        return self._get_coll(
            '_hl_dry_respiration_coll', self._heat_loss_dry_respiration,
            Power('Heat Loss From Dry Respiration'), 'W')

    @property
    def heat_loss_radiation(self):
        """Data Collection of heat loss by radiation in [W]."""
        return self._get_coll('_hl_radiation_coll', self._heat_loss_radiation,
                              Power('Heat Loss From Radiation'), 'W')

    @property
    def heat_loss_convection(self):
        """Data Collection of heat loss by convection in [W]."""
        return self._get_coll('_hl_convection_coll', self._heat_loss_convection,
                              Power('Heat Loss From Convection'), 'W')



# coding=utf-8
"""Parameters for specifying body characteristics for the SolarCal model."""
from __future__ import division

from ._base import ComfortParameter
from ..solarcal import sharp_from_solar_and_body_azimuth


class SolarCalParameter(ComfortParameter):
    """Parameters specifying body characteristics for the SolarCal model.

    Properties:
        posture
        sharp
        body_azimuth
        body_absorptivity
        body_emissivity
        acceptable_postures
    """
    _model = 'SolarCal'
    _acceptable_postures = ('standing', 'seated', 'supine')

    def __init__(self, posture=None, sharp=None, body_azimuth=None,
                 body_absorptivity=None, body_emissivity=None):
        """Initalize SolarCal Body Parameters.

        Args:
            posture: A text string indicating the posture of the body. Letters must
                be lowercase.  Choose from the following: "standing", "seated", "supine".
                Default is "standing".
            sharp: A number between 0 and 180 representing the solar horizontal
                angle relative to front of person (SHARP). 0 signifies sun that is
                shining directly into the person's face and 180 signifies sun that
                is shining at the person's back. Default is 135, asuming that a person
                typically faces their side or back to the sun to avoid glare.
            body_azimuth: A number (between 0 and 360) representing the direction that
                the human is facing in degrees (0=North, 90=East, 180=South, 270=West).
                If this number is greater than 360 or less than 0, it will be converted
                to the correct angle within this range.
                Default is None, which will assume that the sharp input dictates the
                degrees the human is facing from the sun.
            body_absorptivity: A number between 0 and 1 representing the average
                shortwave absorptivity of the body (including clothing and skin color).
                Typical clothing values - white: 0.2, khaki: 0.57, black: 0.88
                Typical skin values - white: 0.57, brown: 0.65, black: 0.84
                Default is 0.7 for average (brown) skin and medium clothing.
            body_emissivity: A number between 0 and 1 representing the average
                longwave emissivity of the body.  Default is 0.95, which is almost
                always the case except in rare situations of wearing metalic clothing.
        """
        if posture is not None:
            assert isinstance(posture, str), 'posture must be a string.'\
                ' Got {}'.format(type(posture))
            assert posture.lower() in self._acceptable_postures, 'posture {} is not '\
                'acceptable. Choose from {}'.format(posture, self._acceptable_postures)
            self._posture = posture
        else:
            self._posture = 'standing'

        if sharp is not None and body_azimuth is not None:
            raise TypeError('sharp and body_azimuth are mutually exclusive.\n'
                            'You may specify one or the other but not both.\n'
                            'Set one of them to None.')
        elif body_azimuth is not None:
            if body_azimuth > 360:
                while body_azimuth > 360:
                    body_azimuth = body_azimuth - 360
            elif body_azimuth < 0:
                while body_azimuth < 0:
                    body_azimuth = body_azimuth + 360
            self._sharp = None
            self._body_azimuth = body_azimuth
        elif sharp is not None:
            assert 0 <= sharp <= 180, 'sharp must be between 0 and 180.'\
                ' Got {}'.format(sharp)
            self._sharp = sharp
            self._body_azimuth = None
        else:
            self._sharp = 135
            self._body_azimuth = None

        if body_absorptivity is not None:
            assert 0 <= body_absorptivity <= 1, 'body_absorptivity must be between'\
                ' 0 and 1. Got {}'.format(body_absorptivity)
            self._body_absorptivity = body_absorptivity
        else:
            self._body_absorptivity = 0.7

        if body_emissivity is not None:
            assert 0 <= body_emissivity <= 1, 'body_emissivity must be between'\
                ' 0 and 1. Got {}'.format(body_emissivity)
            self._body_emissivity = body_emissivity
        else:
            self._body_emissivity = 0.95

    @property
    def posture(self):
        """A text string indicating the posture of the body."""
        return self._posture

    @property
    def sharp(self):
        """The solar horizontal angle relative to front of person (SHARP).

        Between 0 and 180. 0 signifies sun that is shining directly into the person's
        face and 180 signifies sun that is shining at the person's back."""
        return self._sharp

    @property
    def body_azimuth(self):
        """Number representing the direction that the human is facing in degrees.

        Between 0 and 360. 0=North, 90=East, 180=South, 270=West."""
        return self._body_azimuth

    @property
    def body_absorptivity(self):
        """Number representing the average shortwave absorptivity of the body.

        Between 0 and 1. Includes clothing and skin color."""
        return self._body_absorptivity

    @property
    def body_emissivity(self):
        """Number representing the average longwave emissivity of the body.

        Between 0 and 1. Typically 0.95."""
        return self._body_emissivity

    @property
    def acceptable_postures(self):
        """Tuple with acceptable inputs for the posture property."""
        return self._acceptable_postures

    def get_sharp(self, solar_azimuth):
        if self.sharp is not None:
            return self.sharp
        return sharp_from_solar_and_body_azimuth(solar_azimuth, self.body_azimuth)

    def duplicate(self):
        """Duplicate SolarCal Parameters."""
        return SolarCalParameter(self.posture, self.sharp, self.body_azimuth,
                                 self.body_absorptivity, self.body_emissivity)

    def __repr__(self):
        """SolarCal body parameters representation."""
        if self.sharp is not None:
            return "SolarCal Body Parameters\n Posture: {}\n SHARP: {}"\
                "\n Body Absortivity: {}\n Body Emissivity: {}".format(
                    self.posture, self.sharp, self.body_absorptivity,
                    self.body_emissivity)
        else:
            return "SolarCal Body Parameters\n Posture: {}\n Body Azimuth: {}"\
                "\n Body Absortivity: {}\n Body Emissivity: {}".format(
                    self.posture, self.body_azimuth, self.body_absorptivity,
                    self.body_emissivity)



# coding=utf-8
"""Collection of methods that extends current objects in ladybug core library.

This module is imported in __init__.py and applies the changes on runtime.
"""
from __future__ import division

from ladybug.epw import EPW

from .collection.utci import UTCI
from .collection.pmv import PMV
from .collection.solarcal import OutdoorSolarCal


def get_universal_thermal_climate_index(self, include_wind=True, include_sun=True,
                                        utci_parameter=None):
    """Get a UTCI comfort object from the conditions within the EPW file.

    Args:
        include_wind: Set to True to include the EPW wind speed in the calculation.
            Setting to False will assume a condition that is shielded from wind
            where the human expereinces a very low wind speed of 0.1 m/s.
            Default is True to include wind.
        include_sun: Set to True to include the mean radiant temperature (MRT) delta
            from both shortwave solar falling directly on people and long wave radiant
            exchange with the sky. Setting to False will assume a shaded condition
            with MRT being equal to the EPW dry bulb temperature. When set to True,
            this calculation will assume no surrounding shade context, standing human
            geometry, and a solar horizontal angle relative to front of person (SHARP)
            of 135 degrees. A SHARP of 135 essentially assumes that a person typically
            faces their side or back to the sun to avoid glare.
            Default is True to include sun.
        utci_parameter: Optional UTCIParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume comfort thresholds consistent with those used by meterologists
            to categorize outdoor conditions.

    Returns:
        utci_obj: An object with data collections of the UTCI results as properties.
            Properties include (but are not limited to) the following:
                utci
                is_comfortable
                thermal_condition
                thermal_condition_five_point
                thermal_condition_seven_point
                thermal_condition_nine_point
                thermal_condition_eleven_point
                original_utci_category
    """
    # Get wind and mrt inputs
    wind_speed = self.wind_speed if include_wind is True else 0.1
    if include_sun is True:
        solarcal_obj = OutdoorSolarCal(self.location, self.direct_normal_radiation,
                                       self.diffuse_horizontal_radiation,
                                       self.horizontal_infrared_radiation_intensity,
                                       self.dry_bulb_temperature)
        mrt = solarcal_obj.mean_radiant_temperature
    else:
        mrt = self.dry_bulb_temperature

    # return the comfort object
    utci_obj = UTCI(self.dry_bulb_temperature, self.relative_humidity, mrt, wind_speed,
                    utci_parameter)
    return utci_obj


def get_standard_effective_temperature(self, include_wind=True, include_sun=True,
                                       met_rate=None, clo_value=None, external_work=None,
                                       pmv_parameter=None):
    """Get a SET comfort object from the conditions within the EPW file.

    Args:
        include_wind: Set to True to include the EPW wind speed in the calculation.
            Setting to False will assume a condition that is shielded from wind
            where the human expereinces a very low wind speed of 0.1 m/s.
            Default is True to include wind.
        include_sun: Set to True to include the mean radiant temperature (MRT) delta
            from both shortwave solar falling directly on people and long wave radiant
            exchange with the sky. Setting to False will assume a shaded condition
            with MRT being equal to the EPW dry bulb temperature. When set to True,
            this calculation will assume no surrounding shade context, standing human
            geometry, and a solar horizontal angle relative to front of person (SHARP)
            of 135 degrees. A SHARP of 135 essentially assumes that a person typically
            faces their side or back to the sun to avoid glare.
            Default is True to include sun.
        met_rate: Data Collection of metabolic rate in met or a single
            metabolic rate value to be used for the whole analysis. If None,
            default is set to 1.1 met (for seated, typing).
        clo_value: Data Collection of clothing values rate in clo or a single
            clothing value to be used for the whole analysis. If None, default is
            set to 0.7 clo (for long sleeve shirt and pants).
        external_work: Data Collection of external work in met or a single
            external work value to be used for the whole analysis. If None,
            default is set to 0 met.
        pmv_parameter: Optional PMVParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume a PPD threshold of 10%, no absolute humidity constraints
            and a still air threshold of 0.1 m/s.

    Returns:
        set_obj: An object with data collections of the SET results as properties.
            Properties include (but are not limited to) the following:
                standard_effective_temperature
                pmv
                ppd
                is_comfortable
                thermal_condition
    """
    # Get wind and mrt inputs
    if include_wind is True:
        wind_speed = self.wind_speed.duplicate()
        for i, spd in enumerate(wind_speed):
            wind_speed[i] = spd * (2 / 3)
    else:
        wind_speed = 0.1

    if include_sun is True:
        solarcal_obj = OutdoorSolarCal(self.location, self.direct_normal_radiation,
                                       self.diffuse_horizontal_radiation,
                                       self.horizontal_infrared_radiation_intensity,
                                       self.dry_bulb_temperature)
        mrt = solarcal_obj.mean_radiant_temperature
    else:
        mrt = self.dry_bulb_temperature

    # return the comfort object
    set_obj = PMV(self.dry_bulb_temperature, self.relative_humidity, mrt, wind_speed,
                  met_rate, clo_value, external_work, pmv_parameter)
    return set_obj


EPW.get_universal_thermal_climate_index = get_universal_thermal_climate_index
EPW.get_standard_effective_temperature = get_standard_effective_temperature
